# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict
import re
import logging

from glyphsLib.util import bin_to_int_list, int_list_to_bin
from .filters import parse_glyphs_filter
from .common import to_ufo_color
from .constants import (
    GLYPHS_PREFIX,
    UFO2FT_COLOR_PALETTES_KEY,
    UFO2FT_FILTERS_KEY,
    UFO2FT_USE_PROD_NAMES_KEY,
    CODEPAGE_RANGES,
    REVERSE_CODEPAGE_RANGES,
    PUBLIC_PREFIX,
    UFO_FILENAME_CUSTOM_PARAM,
    UFO2FT_META_TABLE_KEY,
)
from .features import replace_feature, replace_prefixes

"""Set Glyphs custom parameters in UFO info or lib, where appropriate.

Custom parameter data will be extracted from a Glyphs object such as GSFont,
GSFontMaster or GSInstance by wrapping it in the GlyphsObjectProxy.
This proxy normalizes and speeds up the API used to access custom parameters,
and also keeps track of which customParameters have been read from the object.

Note:
    In the special case of GSInstance -> UFO, the source object is not
    actually the GSInstance but a designspace InstanceDescriptor wrapped in
    InstanceDescriptorAsGSInstance. This is because the generation of
    instance UFOs from a Glyphs font happens in two steps:

        1. the GSFont is turned into a designspace + master UFOS
        2. the designspace + master UFOs are interpolated into instance UFOs

    We want step 2. to rely only on information from the designspace, that's why
    we use the InstanceDescriptor as a source of customParameters to put into
    the instance UFO.

In the other direction, put information from UFO info or lib into a GSFont or a
GSFontMaster. The UFO source is wrapped in a UFOProxy that records which
attributes are read/written.

In order to go in both directions, each known parameter is managed by a
ParamHandler object that can implement special rules to translate the value
between Glyphs and UFO formats. This files aims at providing at least one
handler per defined UFO info attribute, plus a bunch of handlers for known
Custom Paramerters or known UFO lib elements.

To go for example from UFO to Glyphs, each registered ParamHandler is called,
and each tries to find its parameter in the UFO's info or lib data. Accesses to
the UFO lib are recorded by the UFO proxy. After all registered ParamHandlers
have worked, we know which UFO lib fields have been "consumed" in a smart way,
and we can stupidly copy the other ones over to the Glyphs side. Same when
going from Glyphs to UFOs.
"""

CUSTOM_PARAM_PREFIX = GLYPHS_PREFIX + "customParameter."


logger = logging.getLogger(__name__)


def identity(value):
    return value


class GlyphsObjectProxy:
    """Accelerate and record access to the glyphs object's custom parameters"""

    def __init__(self, glyphs_object, glyphs_module, ignore_disabled=False):
        self._owner = glyphs_object
        # This is a key part to be used in UFO lib keys to be able to choose
        # between master and font attributes during roundtrip
        self.sub_key = glyphs_object.__class__.__name__ + "."
        self._glyphs_module = glyphs_module
        self._lookup = defaultdict(list)
        for param in glyphs_object.customParameters:
            if ignore_disabled and param.disabled:
                continue
            self._lookup[param.name].append(param.value)
        self._handled = set()

    def get_attribute_value(self, key):
        if not hasattr(self._owner, key):
            return None
        return getattr(self._owner, key)

    def set_attribute_value(self, key, value):
        if not hasattr(self._owner, key):
            return
        setattr(self._owner, key, value)

    def get_custom_value(self, key):
        """Return the first and only custom parameter matching the given name."""
        self._handled.add(key)
        values = self._lookup[key]
        if len(values) > 1:
            raise RuntimeError(f"More than one value for this customParameter: {key}")
        if values:
            return values[0]
        return None

    def get_custom_values(self, key):
        """Return a set of values for the given customParameter name."""
        self._handled.add(key)
        return self._lookup[key]

    def set_custom_value(self, key, value):
        """Set one custom parameter with the given value.
        We assume that the list of custom parameters does not already contain
        the given parameter so we only append.
        """
        self._owner.customParameters.append(
            self._glyphs_module.GSCustomParameter(name=key, value=value)
        )

    def set_custom_values(self, key, values):
        """Set several values for the customParameter with the given key.
        We append one GSCustomParameter per value.
        """
        for value in values:
            self.set_custom_value(key, value)

    def unhandled_custom_parameters(self):
        for param in self._owner.customParameters:
            if param.name not in self._handled:
                yield param

    def mark_handled(self, key):
        """Mark a key as handled so it is ignored by `unhandled_custom_parameters`.

        Use e.g. when you handle a custom parameter outside this module.
        """
        self._handled.add(key)

    def is_font(self):
        """Returns whether we are looking at a top-level GSFont object as
        opposed to a master or instance."""
        return hasattr(self._owner, "glyphs")

    def get_property(self, key):
        if key and hasattr(self._owner, "properties"):
            return self._owner.properties.get(key)
        return None


class UFOProxy:
    """Record access to the UFO's lib custom parameters"""

    def __init__(self, ufo):
        self._owner = ufo
        self._handled = set()

    def has_info_attr(self, name):
        return hasattr(self._owner.info, name)

    def get_info_value(self, name):
        return getattr(self._owner.info, name)

    def set_info_value(self, name, value):
        setattr(self._owner.info, name, value)

    def has_lib_key(self, name):
        return name in self._owner.lib

    def get_lib_value(self, name):
        if name not in self._owner.lib:
            return None
        self._handled.add(name)
        return self._owner.lib[name]

    def set_lib_value(self, name, value):
        self._owner.lib[name] = value

    def unhandled_lib_items(self):
        for key, value in self._owner.lib.items():
            if key.startswith(CUSTOM_PARAM_PREFIX) and key not in self._handled:
                yield (key, value)


class AbstractParamHandler:
    # @abstractmethod
    def to_glyphs(self):
        pass

    # @abstractmethod
    def to_ufo(self):
        pass


class ParamHandler(AbstractParamHandler):
    def __init__(
        self,
        glyphs_name,
        ufo_name=None,
        glyphs_long_name=None,
        glyphs_multivalued=False,
        glyphs3_property=None,
        ufo_prefix=CUSTOM_PARAM_PREFIX,
        ufo_info=True,
        ufo_default=None,
        value_to_ufo=identity,
        value_to_glyphs=identity,
    ):
        self.glyphs_name = glyphs_name
        self.glyphs_long_name = glyphs_long_name
        self.glyphs_multivalued = glyphs_multivalued
        self.glyphs3_property = glyphs3_property
        # By default, they have the same name in both
        self.ufo_name = ufo_name or glyphs_name
        self.ufo_prefix = ufo_prefix
        self.ufo_info = ufo_info
        self.ufo_default = ufo_default
        # Value transformation functions
        self.value_to_ufo = value_to_ufo
        self.value_to_glyphs = value_to_glyphs

    # By default, the parameter is read from/written to:
    #  - the Glyphs object's customParameters
    #  - the UFO's info object if it has a matching attribute, else the lib
    def to_glyphs(self, glyphs, ufo):
        ufo_value = self._read_from_ufo(glyphs, ufo)
        if ufo_value is None:
            return
        glyphs_value = self.value_to_glyphs(ufo_value)
        self._write_to_glyphs(glyphs, glyphs_value)

    def to_ufo(self, builder, glyphs, ufo):
        glyphs_value = self._read_from_glyphs(glyphs)
        if glyphs_value is None:
            return
        ufo_value = self.value_to_ufo(glyphs_value)
        self._write_to_ufo(glyphs, ufo, ufo_value)

    def _read_from_glyphs(self, glyphs):
        value = None
        # Try to read from the properties first.
        value = glyphs.get_property(self.glyphs3_property)
        if value is not None:
            return value
        # Try to read from the custom parameters next, using both the short
        # name, which has precedence, and the prefixed (long) name.
        if self.glyphs_multivalued:
            getter = glyphs.get_custom_values
        else:
            getter = glyphs.get_custom_value
        value = getter(self.glyphs_name)
        if value is not None:
            return value
        if self.glyphs_long_name is not None:
            value = getter(self.glyphs_long_name)
        return value

    def _write_to_glyphs(self, glyphs, value):
        # We currently convert UFO to Glyphs2 files.
        # If we ever export Glyphs3 by default, we need a similar test
        # here to the one in _read_from_glyphs to determine whether a
        # value should be placed in the new properties top-level key.

        # Never write the prefixed (long) name?
        # FIXME: (jany) maybe should rather preserve the naming choice of user
        if self.glyphs_multivalued:
            glyphs.set_custom_values(self.glyphs_name, value)
        else:
            glyphs.set_custom_value(self.glyphs_name, value)

    def _read_from_ufo(self, glyphs, ufo):
        if self.ufo_info and ufo.has_info_attr(self.ufo_name):
            return ufo.get_info_value(self.ufo_name)
        else:
            ufo_prefix = self.ufo_prefix
            if ufo_prefix == CUSTOM_PARAM_PREFIX:
                ufo_prefix += glyphs.sub_key
            return ufo.get_lib_value(ufo_prefix + self.ufo_name)

    def _write_to_ufo(self, glyphs, ufo, value):
        if self.ufo_default is not None and value == self.ufo_default:
            return
        if self.ufo_info and ufo.has_info_attr(self.ufo_name):
            # most OpenType table entries go in the info object
            ufo.set_info_value(self.ufo_name, value)
        else:
            # everything else gets dumped in the lib
            ufo_prefix = self.ufo_prefix
            if ufo_prefix == CUSTOM_PARAM_PREFIX:
                ufo_prefix += glyphs.sub_key
            ufo.set_lib_value(ufo_prefix + self.ufo_name, value)


KNOWN_PARAM_HANDLERS = []


def register(handler):
    KNOWN_PARAM_HANDLERS.append(handler)


GLYPHS_UFO_CUSTOM_PARAMS = (
    # These are be stored in the official descriptor attributes.
    # "familyName",
    # "fileName",
    ("compatibleFullName", "openTypeNameCompatibleFullName"),
    ("hheaAscender", "openTypeHheaAscender"),
    ("hheaDescender", "openTypeHheaDescender"),
    ("hheaLineGap", "openTypeHheaLineGap"),
    # OS/2 parameters
    ("panose", "openTypeOS2Panose"),
    ("fsType", "openTypeOS2Type"),
    ("typoAscender", "openTypeOS2TypoAscender"),
    ("typoDescender", "openTypeOS2TypoDescender"),
    ("typoLineGap", "openTypeOS2TypoLineGap"),
    ("unicodeRanges", "openTypeOS2UnicodeRanges"),
    ("strikeoutSize", "openTypeOS2StrikeoutSize"),
    ("strikeoutPosition", "openTypeOS2StrikeoutPosition"),
    # OS/2 subscript parameters
    ("subscriptXSize", "openTypeOS2SubscriptXSize"),
    ("subscriptYSize", "openTypeOS2SubscriptYSize"),
    ("subscriptXOffset", "openTypeOS2SubscriptXOffset"),
    ("subscriptYOffset", "openTypeOS2SubscriptYOffset"),
    # OS/2 superscript parameters
    ("superscriptXSize", "openTypeOS2SuperscriptXSize"),
    ("superscriptYSize", "openTypeOS2SuperscriptYSize"),
    ("superscriptXOffset", "openTypeOS2SuperscriptXOffset"),
    ("superscriptYOffset", "openTypeOS2SuperscriptYOffset"),
    # These can be recovered by reading the mapping backward.
    # ("weightClass", "openTypeOS2WeightClass"),
    # ("widthClass", "openTypeOS2WidthClass"),
    # These are processed separatedly down below.
    # ("winAscent", "openTypeOS2WinAscent"),
    # ("winDescent", "openTypeOS2WinDescent"),
    ("vheaVertAscender", "openTypeVheaVertTypoAscender"),
    ("vheaVertDescender", "openTypeVheaVertTypoDescender"),
    ("vheaVertLineGap", "openTypeVheaVertTypoLineGap"),
    ("vheaVertTypoAscender", "openTypeVheaVertTypoAscender"),
    ("vheaVertTypoDescender", "openTypeVheaVertTypoDescender"),
    ("vheaVertTypoLineGap", "openTypeVheaVertTypoLineGap"),
    # PostScript parameters
    ("blueScale", "postscriptBlueScale"),
    ("blueShift", "postscriptBlueShift"),
    ("isFixedPitch", "postscriptIsFixedPitch"),
    ("underlinePosition", "postscriptUnderlinePosition"),
    ("underlineThickness", "postscriptUnderlineThickness"),
)
for glyphs_name, ufo_name in GLYPHS_UFO_CUSTOM_PARAMS:
    register(ParamHandler(glyphs_name, ufo_name, glyphs_long_name=ufo_name))

# Reference:
# https://github.com/googlefonts/glyphsLib/pull/881#issuecomment-1474226616
GLYPHS_UFO_CUSTOM_PARAMS_GLYPHS3_PROPERTIES = (
    # This is stored in the official descriptor attributes.
    # "familyNames",
    # TODO: Map these properties to custom parameters if applicable.
    # "designers",
    # "designerURL",
    # "manufacturers",
    # "manufacturerURL",
    # "copyrights",
    ("versionString", "openTypeNameVersion", "versionString"),
    ("vendorID", "openTypeOS2VendorID", "vendorID"),
    ("uniqueID", "openTypeNameUniqueID", "uniqueID"),
    ("license", "openTypeNameLicense", "licenses"),
    ("licenseURL", "openTypeNameLicenseURL", "licenseURL"),
    ("trademark", "trademark", "trademarks"),
    ("description", "openTypeNameDescription", "descriptions"),
    ("sampleText", "openTypeNameSampleText", "sampleTexts"),
    # TODO: Should the postscriptFullName or postscriptFullNames property be
    # used for the postscriptFullName custom parameter?
    # "postscriptFullNames",
    ("postscriptFullName", "postscriptFullName", "postscriptFullName"),
    # TODO: This is stored in the official descriptor attibutes. Should this
    # entry be removed?
    ("postscriptFontName", "postscriptFontName", "postscriptFontName"),
    # TODO: Map these properties to custom parameters if applicable.
    # "compatibleFullNames",
    # "styleNames",
    # "styleMapFamilyNames",
    # "styleMapStyleNames",
    ("preferredFamilyName", "openTypeNamePreferredFamilyName", "preferredFamilyNames"),
    (
        "preferredSubfamilyName",
        "openTypeNamePreferredSubfamilyName",
        "preferredSubfamilyNames",
    ),
    # TODO: Map this property to a custom parameter if applicable.
    # "variableStyleNames",
    ("WWSFamilyName", "openTypeNameWWSFamilyName", "WWSFamilyName"),
    ("WWSSubfamilyName", "openTypeNameWWSSubfamilyName", "WWSSubfamilyName"),
    # TODO: Map this property to a custom parameter if applicable.
    # "variationsPostScriptNamePrefix",
)

for glyphs_name, ufo_name, property_name in GLYPHS_UFO_CUSTOM_PARAMS_GLYPHS3_PROPERTIES:
    register(
        ParamHandler(
            glyphs_name,
            ufo_name,
            glyphs_long_name=ufo_name,
            glyphs3_property=property_name,
        )
    )

# TODO: (jany) for all the following fields, check that they are stored in a
# meaningful Glyphs customParameter. Maybe they have short names?
GLYPHS_UFO_CUSTOM_PARAMS_NO_SHORT_NAME = (
    "openTypeHheaCaretSlopeRun",
    "openTypeVheaCaretSlopeRun",
    "openTypeHheaCaretSlopeRise",
    "openTypeVheaCaretSlopeRise",
    "openTypeHheaCaretOffset",
    "openTypeVheaCaretOffset",
    "openTypeHeadLowestRecPPEM",
    "openTypeHeadFlags",
    "openTypeOS2FamilyClass",
    "postscriptSlantAngle",
    "postscriptUniqueID",
    # TODO: Should this be handled in `blue_values.py`?
    # "postscriptFamilyBlues",
    # "postscriptFamilyOtherBlues",
    "postscriptBlueFuzz",
    "postscriptForceBold",
    "postscriptDefaultWidthX",
    "postscriptNominalWidthX",
    "postscriptWeightName",
    "postscriptDefaultCharacter",
    "postscriptWindowsCharacterSet",
    "macintoshFONDFamilyID",
    "macintoshFONDName",
    "styleMapFamilyName",
    "styleMapStyleName",
)
for name in GLYPHS_UFO_CUSTOM_PARAMS_NO_SHORT_NAME:
    register(ParamHandler(name))


class EmptyListDefaultParamHandler(ParamHandler):
    def to_glyphs(self, glyphs, ufo):
        ufo_value = self._read_from_ufo(glyphs, ufo)
        # Ingore default value == empty list
        if ufo_value is None or ufo_value == []:
            return
        glyphs_value = self.value_to_glyphs(ufo_value)
        self._write_to_glyphs(glyphs, glyphs_value)


register(EmptyListDefaultParamHandler("postscriptFamilyBlues"))
register(EmptyListDefaultParamHandler("postscriptFamilyOtherBlues"))


# Convert code page numbers to OS/2 ulCodePageRange bits. Empty lists stay empty lists.
class OS2CodePageRangesParamHandler(AbstractParamHandler):
    def to_glyphs(self, glyphs, ufo):
        ufo_codepage_bits = ufo.get_info_value("openTypeOS2CodePageRanges")
        if ufo_codepage_bits is None:
            return

        codepages = []
        unsupported_codepage_bits = []
        for codepage in ufo_codepage_bits:
            if codepage in REVERSE_CODEPAGE_RANGES:
                codepages.append(REVERSE_CODEPAGE_RANGES[codepage])
            else:
                unsupported_codepage_bits.append(codepage)

        glyphs.set_custom_value("codePageRanges", codepages)
        if unsupported_codepage_bits:
            glyphs.set_custom_value(
                "codePageRangesUnsupportedBits", unsupported_codepage_bits
            )

    def to_ufo(self, builder, glyphs, ufo):
        codepages = glyphs.get_custom_value("codePageRanges")
        if codepages is None:
            codepages = glyphs.get_custom_value("openTypeOS2CodePageRanges")
            if codepages is None:
                return

        ufo_codepage_bits = self._convert_to_bits(codepages)
        unsupported_codepage_bits = glyphs.get_custom_value(
            "codePageRangesUnsupportedBits"
        )
        if unsupported_codepage_bits:
            ufo_codepage_bits.extend(unsupported_codepage_bits)

        ufo.set_info_value("openTypeOS2CodePageRanges", sorted(ufo_codepage_bits))

    @staticmethod
    def _convert_to_bits(codepages):
        ufo_codepage_bits = []
        for codepage in codepages:
            if isinstance(codepage, int):
                ufo_codepage_bits.append(CODEPAGE_RANGES[codepage])

            elif isinstance(codepage, str):
                if codepage.isdigit():
                    ufo_codepage_bits.append(CODEPAGE_RANGES[int(codepage)])
                elif codepage.startswith("bit "):
                    try:
                        ufo_codepage_bits.append(int(codepage[4:]))
                    except ValueError as err:
                        raise ValueError(
                            f"'{codepage}' is not in correct format. "
                            "A number must follow after 'bit '."
                        ) from err
                else:
                    raise ValueError(
                        f"'{codepage}'is neither a number nor 'bit ' format."
                    )

            else:
                raise TypeError(f"Unsupported type: {type(codepage)}")

        return ufo_codepage_bits


register(OS2CodePageRangesParamHandler())

# enforce that winAscent/Descent are positive, according to UFO spec
for glyphs_name in ("winAscent", "winDescent"):
    ufo_name = "openTypeOS2W" + glyphs_name[1:]
    register(
        ParamHandler(
            glyphs_name,
            ufo_name,
            glyphs_long_name=ufo_name,
            value_to_ufo=abs,
            value_to_glyphs=abs,
        )
    )

# The value of these could be a float, and ufoLib/defcon expect an int.
for glyphs_name in ("weightClass", "widthClass"):
    ufo_name = "openTypeOS2W" + glyphs_name[1:]
    register(ParamHandler(glyphs_name, ufo_name, value_to_ufo=int))


# convert Glyphs' GASP Table to UFO openTypeGaspRangeRecords
def to_ufo_gasp_table(value):
    # XXX maybe the parser should cast the gasp values to int?
    value = {int(k): int(v) for k, v in value.items()}
    gasp_records = []
    # gasp range records must be sorted in ascending rangeMaxPPEM
    for max_ppem, gasp_behavior in sorted(value.items()):
        gasp_records.append(
            {
                "rangeMaxPPEM": max_ppem,
                "rangeGaspBehavior": bin_to_int_list(gasp_behavior),
            }
        )
    return gasp_records


def to_glyphs_gasp_table(value):
    return {
        str(record["rangeMaxPPEM"]): int_list_to_bin(record["rangeGaspBehavior"])
        for record in value
    }


register(
    ParamHandler(
        glyphs_name="GASP Table",
        ufo_name="openTypeGaspRangeRecords",
        value_to_ufo=to_ufo_gasp_table,
        value_to_glyphs=to_glyphs_gasp_table,
    )
)

register(
    ParamHandler(
        glyphs_name="gasp Table",
        ufo_name="openTypeGaspRangeRecords",
        value_to_ufo=to_ufo_gasp_table,
        value_to_glyphs=to_glyphs_gasp_table,
    )
)


# convert Glyphs' meta Table to UFO openTypeMeta
def to_ufo_meta_table(value):
    meta = {}
    # In:  {data = "de-Latn"; tag = dlng; }, {data = "sr-Cyrl"; tag = slng; }
    # Out: { "dlng": [ "de-Latn" ], "slng": [ "sr-Cyrl" ] }
    for entry in value:
        tag, data = entry["tag"], entry["data"]
        if tag in meta:
            logger.warning(
                f"Multiple '{tag}' tags in meta table; only the last one will be used"
            )

        if tag in ("appl", "bild"):
            meta[tag] = data
        else:
            meta.setdefault(tag, []).append(data)
    return meta


def to_glyphs_meta_table(value):
    meta = []
    for tag, data in value.items():
        if isinstance(data, list):
            for entry in data:
                meta.append({"tag": tag, "data": entry})
        else:
            meta.append({"tag": tag, "data": data})
    return meta


register(
    ParamHandler(
        glyphs_name="meta Table",
        ufo_name=UFO2FT_META_TABLE_KEY,
        ufo_info=False,
        ufo_prefix="",
        value_to_ufo=to_ufo_meta_table,
        value_to_glyphs=to_glyphs_meta_table,
    )
)


def to_ufo_color_palettes(value):
    return [[to_ufo_color(color) for color in palette] for palette in value]


def _to_glyphs_color(color):
    if color[0] == color[1] == color[2]:
        color = [color[0], color[3]]
    return ",".join(str(round(v * 255)) for v in color)


def to_glyphs_color_palettes(value):
    return [[_to_glyphs_color(color) for color in palette] for palette in value]


register(
    ParamHandler(
        glyphs_name="Color Palettes",
        ufo_name=UFO2FT_COLOR_PALETTES_KEY,
        ufo_info=False,
        ufo_prefix="",
        value_to_ufo=to_ufo_color_palettes,
        value_to_glyphs=to_glyphs_color_palettes,
    )
)


class NameRecordParamHandler(AbstractParamHandler):
    def to_entry(self, record):
        identifiers = [
            record["nameID"],
            record["platformID"],
            record["encodingID"],
            record["languageID"],
        ]
        encoding = " ".join(map(str, identifiers))
        string = record["string"]

        return f"{encoding}; {string}"

    def parse_decimal(self, string):
        # In Python octal strings must start with a prefix. Glyphs
        # uses AFDKO decimal number specification which allows
        # octals starting with "0".
        if string.startswith("0x"):
            return int(string, 16)
        elif string.startswith("0"):
            return int(string, 8)
        else:
            return int(string, 10)

    # See the Glyphs manual for the Name Table Entry format:
    # https://glyphsapp.com/media/pages/learn/3ec528a11c-1634835554/glyphs-3-handbook.pdf
    def to_record(self, entry):
        # Split only on the first semicolon occurance. Glyphs doesn't
        # have any special escaping, so anything after the first
        # semicolon is treated as part of the name table entry.
        parts = entry.split(";", 1)

        if len(parts) != 2:
            logger.warning(f"Invalid Name Table Entry '{entry}' ignored.")
        else:
            identifiers = parts[0].split(" ")
            # Strip whitespace. This behaviour is undefined, but
            # it seems sensible to remove leading and trailing spaces.
            string = parts[1].strip()

            try:
                name_id = self.parse_decimal(identifiers[0])
                platform_id = 3  # Windows
                encoding_id = 1  # Unicode BMP
                language_id = 0x409  # English, United States

                if len(identifiers) >= 2:
                    platform_id = self.parse_decimal(identifiers[1])

                if len(identifiers) >= 3:
                    encoding_id = self.parse_decimal(identifiers[2])
                elif platform_id == 1:
                    encoding_id = 0

                if len(identifiers) >= 4:
                    language_id = self.parse_decimal(identifiers[3])
                elif platform_id == 1:
                    language_id = 0

                return {
                    "nameID": name_id,
                    "platformID": platform_id,
                    "encodingID": encoding_id,
                    "languageID": language_id,
                    "string": string,
                }
            except ValueError:
                logger.warning(f"Invalid name table identifiers '{parts[0]}'.")

    def to_glyphs(self, glyphs, ufo):
        if glyphs.is_font():
            records = ufo.get_info_value("openTypeNameRecords")
            if records:
                entries = [self.to_entry(record) for record in records]

                glyphs.set_custom_values("Name Table Entry", entries)

    def to_ufo(self, builder, glyphs, ufo):
        entries = glyphs.get_custom_values("Name Table Entry")
        if entries:
            records = ufo.get_info_value("openTypeNameRecords") or []
            for entry in entries:
                record = self.to_record(entry)
                if record is not None:
                    records.append(record)
            ufo.set_info_value("openTypeNameRecords", records)


register(NameRecordParamHandler())


register(ParamHandler(glyphs_name="Disable Last Change", ufo_name="disablesLastChange"))

register(
    ParamHandler(
        # convert between Glyphs.app's and ufo2ft's equivalent parameter
        glyphs_name="Don't use Production Names",
        ufo_name=UFO2FT_USE_PROD_NAMES_KEY,
        ufo_prefix="",
        value_to_ufo=lambda value: not value,
        value_to_glyphs=lambda value: not value,
    )
)


class MiscParamHandler(ParamHandler):
    """Copy GSFont attributes to ufo lib"""

    def _read_from_glyphs(self, glyphs):
        return glyphs.get_attribute_value(self.glyphs_name)

    def _write_to_glyphs(self, glyphs, value):
        glyphs.set_attribute_value(self.glyphs_name, value)


register(MiscParamHandler(glyphs_name="disablesAutomaticAlignment"))
register(MiscParamHandler(glyphs_name="iconName"))


class DisplayStringsParamHandler(MiscParamHandler):
    def __init__(self):
        super().__init__(glyphs_name="DisplayStrings")

    def to_ufo(self, builder, glyphs, ufo):
        # We test for builder here because apply_instance_data() passes None and
        # we don't want to copy-paste or subclass UFOBuilder.
        if (
            builder is not None
            and builder.store_editor_state
            and builder.font.DisplayStrings
        ):
            super().to_ufo(builder, glyphs, ufo)


register(DisplayStringsParamHandler())

# deal with any Glyphs naming quirks here
register(
    MiscParamHandler(
        glyphs_name="disablesNiceNames",
        ufo_name="useNiceNames",
        value_to_ufo=lambda value: int(not value),
        value_to_glyphs=lambda value: not bool(value),
    )
)

for number in ("", "1", "2", "3"):
    register(MiscParamHandler("customValue" + number, ufo_info=False))
register(MiscParamHandler("weightValue", ufo_info=False))
register(MiscParamHandler("widthValue", ufo_info=False))


def append_unique(array, value):
    if value not in array:
        array.append(value)


class OS2SelectionParamHandler(AbstractParamHandler):
    flags = {7: "Use Typo Metrics", 8: "Has WWS Names"}

    # Note that en empty openTypeOS2Selection list should stay an empty list, as
    # opposed to a non-existant list. In the latter case, we round-trip nothing, in the
    # former, we at least write an empty list to openTypeOS2SelectionUnsupportedBits
    # which we use to re-instate an empty list in the UFO on tripping back.
    def to_glyphs(self, glyphs, ufo):
        ufo_flags = ufo.get_info_value("openTypeOS2Selection")
        if ufo_flags is None:
            return

        unsupported_bits = []
        for flag in ufo_flags:
            if flag in self.flags:
                glyphs.set_custom_value(self.flags[flag], True)
            else:
                unsupported_bits.append(flag)
        glyphs.set_custom_value("openTypeOS2SelectionUnsupportedBits", unsupported_bits)

    def to_ufo(self, builder, glyphs, ufo):
        use_typo_metrics = glyphs.get_custom_value(self.flags[7])
        has_wws_name = glyphs.get_custom_value(self.flags[8])
        unsupported_bits = glyphs.get_custom_value(
            "openTypeOS2SelectionUnsupportedBits"
        )
        if not use_typo_metrics and not has_wws_name and unsupported_bits is None:
            return

        selection_bits = ufo.get_info_value("openTypeOS2Selection") or []
        if use_typo_metrics:
            selection_bits.append(7)
        if has_wws_name:
            selection_bits.append(8)
        if unsupported_bits:
            selection_bits.extend(unsupported_bits)
        ufo.set_info_value("openTypeOS2Selection", sorted(set(selection_bits)))


register(OS2SelectionParamHandler())


class GlyphOrderParamHandler(AbstractParamHandler):
    """Translate between Glyphs.app's glyphOrder parameter and UFO's
    public.glyphOrder.

    See the GlyphOrderTest class for a thorough explanation.
    """

    def to_glyphs(self, glyphs, ufo):
        if glyphs.is_font():
            ufo_glyphOrder = ufo.get_lib_value(PUBLIC_PREFIX + "glyphOrder")
            if ufo_glyphOrder:
                glyphs.set_custom_value("glyphOrder", ufo_glyphOrder)

    def to_ufo(self, builder, glyphs, ufo):
        if glyphs.is_font():
            glyphs_glyphOrder = glyphs.get_custom_value("glyphOrder")
            if glyphs_glyphOrder:
                ufo_glyphOrder = ufo.get_lib_value(PUBLIC_PREFIX + "glyphOrder")
                # If the custom parameter provides partial coverage we want to
                # append the original glyph order for uncovered glyphs.
                glyphs_glyphOrder += [
                    g for g in ufo_glyphOrder if g not in glyphs_glyphOrder
                ]
                ufo.set_lib_value(PUBLIC_PREFIX + "glyphOrder", glyphs_glyphOrder)


register(GlyphOrderParamHandler())


class FilterParamHandler(AbstractParamHandler):
    """Handler for (Pre)Filter custom paramters.

    This is complicated. ufo2ft grew filter modules to mimic some of Glyph's
    automatic features, but due to the impendance mismatch between the flow of
    data in Glyphs and in UFOs plus Designspaces, they need to be handled in
    two ways: once for filters that should be applied to masters and once for
    filters on instances, which should be applied only to interpolated UFOs:

       +------+
       |GSFont+-------------------+
       +----+-+                   |
            |                     |
          +-+-----------+       +-+----------+
          |GSFontMaster |       |GSIntance   |
          +-------------+       +------------+
           userData                    customParameters
             com...ufo2ft.filters        Filter & PreFilter

                ^  |                      |  ^
     roundtrips |  |                      |  |
                |  v                      |  |
            lib                           |  | roundtrips
              com...ufo2ft.filters        |  |
          +-----------+                   v  |
          |Master UFO |          lib
          +---+-------+            com.schriftgestaltung.customParameter...
              |
          +---+-----+        +----------+                    +-----------------+
          | Source  |        | Instance |    ------------>   |Interpolated UFO |
          +---+-----+        +-----+----+                    +-----------------+
              |                    |          goes 1 way        lib
      +-------+-----+              |     apply_instance_data()    com...ufo2ft.filters
      | Designspace +--------------+
      +-------------+

    The ufo2ft filters should roundtrip as-is between UFO source masters and
    GSFontMaster, because that's how we use them in the UFO workflow with 1
    master UFO = 1 final font with filters applied.

    The Glyphs filters defined on GSInstance should keep doing what they were
    doing already:

    - first be copied as-is into the designspace instance's lib, which should
      roundtrip back to Glyphs
    - then be converted to ufo2ft equivalents and put in the final interpolated
      UFOs before they are compiled into final fonts. Those should not
      roundtrip because the interpolated UFO is discarded after compilation.

    The handler below only handles the latter, one-way case. Since ufo2ft
    filters are a UFO lib key, they are automatically stored in a master's
    userData by another code path.
    """

    def to_glyphs(self, glyphs, ufo):
        pass

    def to_ufo(self, builder, glyphs, ufo):
        if not glyphs.is_font():
            ufo_filters = []
            for pre_filter in glyphs.get_custom_values("PreFilter"):
                ufo_filters.append(parse_glyphs_filter(pre_filter, is_pre=True))
            for filter in glyphs.get_custom_values("Filter"):
                ufo_filters.append(parse_glyphs_filter(filter, is_pre=False))

            if not ufo_filters:
                return
            if not ufo.has_lib_key(UFO2FT_FILTERS_KEY):
                ufo.set_lib_value(UFO2FT_FILTERS_KEY, [])
            existing = ufo.get_lib_value(UFO2FT_FILTERS_KEY)
            existing.extend(ufo_filters)


register(FilterParamHandler())


class ReplacePrefixParamHandler(AbstractParamHandler):
    def to_ufo(self, builder, glyphs, ufo):
        repl_map = {}
        for value in glyphs.get_custom_values("Replace Prefix"):
            prefix_name, prefix_code = re.split(r"\s*;\s*", value, maxsplit=1)
            # if multiple 'Replace Prefix' custom params replace the same
            # prefix, the last wins
            repl_map[prefix_name] = prefix_code

        features_text = ufo._owner.features.text

        if not (repl_map and features_text):
            return

        glyph_names = set(ufo._owner.keys())

        ufo._owner.features.text = replace_prefixes(
            repl_map, features_text, glyph_names=glyph_names
        )

    def to_glyphs(self, glyphs, ufo):
        # do the same as ReplaceFeatureParamHandler.to_glyphs
        pass


register(ReplacePrefixParamHandler())


class ReplaceFeatureParamHandler(AbstractParamHandler):
    def to_ufo(self, builder, glyphs, ufo):
        for value in glyphs.get_custom_values("Replace Feature"):
            tag, repl = re.split(r"\s*;\s*", value, maxsplit=1)
            ufo._owner.features.text = replace_feature(
                tag, repl, ufo._owner.features.text or ""
            )

    def to_glyphs(self, glyphs, ufo):
        # TODO: (jany) The "Replace Feature" custom parameter can be used to
        # have one master/instance with different features than what is stored
        # in the GSFont. When going from several UFOs to one GSFont, we could
        # detect when UFOs have different features, put the common ones in
        # GSFont and replace the different ones with this custom parameter.
        # See the file `tests/builder/features_test.py`.
        pass


register(ReplaceFeatureParamHandler())


class ReencodeGlyphsParamHandler(AbstractParamHandler):
    """The "Reencode Glyphs" custom parameter contains a list of
    'glyphname=unicodevalue' strings: e.g., ["smiley=E100", "logo=E101"].
    It only applies to specific instance (not to master or globally) and is
    meant to assign Unicode values to glyphs with the specied name at export
    time.
    When the Unicode value in question is already assigned to another glyph,
    the latter's Unicode value is deleted.
    When the Unicode value is left out, e.g., "f_f_i=", "f_f_j=", this will
    strip "f_f_i" and "f_f_j" of their Unicode values.

    This parameter handler only handles going from Glyphs to (instance) UFOs,
    and not also in the opposite direction, as the parameter isn't stored in
    the UFO lib, but directly applied to the UFO unicode values.
    """

    def to_ufo(self, builder, glyphs, ufo):
        # TODO Check that the wrapped glyphs object is indeed an instance, and
        # not a GSFont or GSMaster (unlikely)
        reencode_list = glyphs.get_custom_value("Reencode Glyphs")
        if not reencode_list:
            return
        ufo = ufo._owner
        cmap = {glyph.unicode: glyph.name for glyph in ufo}
        for entry in reencode_list:
            name, hexcode = entry.split("=")
            if name not in ufo:
                continue
            if hexcode.strip() == "":
                ufo[name].unicode = None
            else:
                codepoint = int(hexcode, 16)
                if codepoint in cmap:
                    previous = cmap[codepoint]
                    ufo[previous].unicode = None
                ufo[name].unicode = codepoint

    def to_glyphs(self, glyphs, ufo):
        # The 'Reencode Glyphs' parameter only applies to instances, which
        # are not meant to be roundtripped. No need to handle it here.
        pass


register(ReencodeGlyphsParamHandler())


class RenameGlyphsParamHandler(AbstractParamHandler):
    """The "Rename Glyphs" custom parameter contains a list of
    'glyphname=glyphname' strings: e.g., ["a=b", "b=a"].
    It only applies to specific instance (not to master or globally).

    The glyph data is swapped, but the unicode assignments remain the
    same.
    """

    def to_ufo(self, builder, glyphs, ufo):
        rename_list = glyphs.get_custom_value("Rename Glyphs")
        if not rename_list:
            return
        ufo = ufo._owner
        for entry in rename_list:
            oldname, newname = entry.split("=")
            ufo[newname], ufo[oldname] = ufo[oldname], ufo[newname]
            ufo[newname].unicodes, ufo[oldname].unicodes = (
                ufo[oldname].unicodes,
                ufo[newname].unicodes,
            )

    def to_glyphs(self, glyphs, ufo):
        # The 'Reencode Glyphs' parameter only applies to instances, which
        # are not meant to be roundtripped. No need to handle it here.
        pass


register(RenameGlyphsParamHandler())


def to_ufo_custom_params(self, ufo, glyphs_object, set_default_params=True):
    # glyphs_module=None because we shouldn't instanciate any Glyphs classes.

    # In 'minimal' mode (enabled e.g. by fontmake when converting .glyphs => .ufo
    # for compiling OpenType fonts) we treat disabled custom parameters as if
    # they are absent.
    # Note that the custom parameters' 'disabled' status is still not saved in the
    # UFO so converting back to .glyphs will mark them as enabled.
    # https://github.com/googlefonts/glyphsLib/issues/905

    # `self` is normally a UFOBuilder instance, but it can be None when this
    # is called by `glyphsLib.builder.instances.apply_instance_data_to_ufo`.
    # The latter function is only ever used by fontmake (to apply instance custom
    # parameters to interpolated instance UFOs) and never in glyphsLib's own
    # glyphs=>ufo=>glyphs code, therefore we assume that when self is None,
    # a 'minimal' build is desired.
    ignore_disabled = getattr(self, "minimal", True)
    glyphs_proxy = GlyphsObjectProxy(
        glyphs_object, glyphs_module=None, ignore_disabled=ignore_disabled
    )
    ufo_proxy = UFOProxy(ufo)

    glyphs_proxy.mark_handled(UFO_FILENAME_CUSTOM_PARAM)

    for handler in KNOWN_PARAM_HANDLERS:
        handler.to_ufo(self, glyphs_proxy, ufo_proxy)

    for param in glyphs_proxy.unhandled_custom_parameters():
        name = _normalize_custom_param_name(param.name)
        ufo.lib[CUSTOM_PARAM_PREFIX + glyphs_proxy.sub_key + name] = param.value

    if set_default_params:
        _set_default_params(ufo)


def to_glyphs_custom_params(self, ufo, glyphs_object):
    glyphs_proxy = GlyphsObjectProxy(glyphs_object, glyphs_module=self.glyphs_module)
    ufo_proxy = UFOProxy(ufo)

    # Handle known parameters
    for handler in KNOWN_PARAM_HANDLERS:
        handler.to_glyphs(glyphs_proxy, ufo_proxy)

    # Since all UFO `info` entries (from `fontinfo.plist`) have a registered
    # handler, the only place where we can find unexpected stuff is the `lib`.
    # See the file `tests/builder/fontinfo_test.py` for `fontinfo` coverage.
    prefix = CUSTOM_PARAM_PREFIX + glyphs_proxy.sub_key
    for name, value in ufo_proxy.unhandled_lib_items():
        name = _normalize_custom_param_name(name)
        if not name.startswith(prefix):
            continue
        name = name[len(prefix) :]
        glyphs_proxy.set_custom_value(name, value)

    _unset_default_params(glyphs_object)


def _normalize_custom_param_name(name):
    """Replace curved quotes with straight quotes in a custom parameter name.
    These should be the only keys with problematic (non-ascii) characters,
    since they can be user-generated.
    """

    replacements = (("\u2018", "'"), ("\u2019", "'"), ("\u201c", '"'), ("\u201d", '"'))
    for orig, replacement in replacements:
        name = name.replace(orig, replacement)
    return name


DEFAULT_PARAMETERS = (
    # ufo2ft defaults to fsType Bit 2 ("Preview & Print embedding"), while
    # Glyphs.app defaults to Bit 3 ("Editable embedding")
    ("fsType", "openTypeOS2Type", [3]),
    # Reference:
    # https://glyphsapp.com/content/1-get-started/2-manuals/
    # 1-handbook-glyphs-2-0/Glyphs-Handbook-2.3.pdf#page=200
    ("underlineThickness", "postscriptUnderlineThickness", 50),
    ("underlinePosition", "postscriptUnderlinePosition", -100),
)


def _set_default_params(ufo):
    """Set Glyphs.app's default parameters when different from ufo2ft ones."""
    for _, ufo_name, default_value in DEFAULT_PARAMETERS:
        if getattr(ufo.info, ufo_name) is None:
            if isinstance(default_value, list):
                # Prevent problem if the same default value list is put in
                # several unrelated objects.
                default_value = default_value[:]
            setattr(ufo.info, ufo_name, default_value)


def _unset_default_params(glyphs):
    """Unset Glyphs.app's parameters that have default values.
    FIXME: (jany) maybe this should be taken care of in the writer? and/or
        classes should have better default values?
    """
    for glyphs_name, _, default_value in DEFAULT_PARAMETERS:
        if (
            glyphs_name in glyphs.customParameters
            and glyphs.customParameters[glyphs_name] == default_value
        ):
            del glyphs.customParameters[glyphs_name]
        # These parameters can be referred to with the two names in Glyphs
        if (
            glyphs_name in glyphs.customParameters
            and glyphs.customParameters[glyphs_name] == default_value
        ):
            del glyphs.customParameters[glyphs_name]


class GSFontParamHandler(ParamHandler):
    def to_glyphs(self, glyphs, ufo):
        if not glyphs.is_font():
            return
        super().to_glyphs(glyphs, ufo)

    def to_ufo(self, builder, glyphs, ufo):
        if not glyphs.is_font():
            return
        super().to_ufo(builder, glyphs, ufo)


# 'Virtual Master' params are GSFont-only and multi-valued (i.e. there can be multiple
# custom parameters named 'Virtual Master'); we know we want them stored in lib.plist
# hence ufo_info=False
register(
    GSFontParamHandler(
        "Virtual Master", ufo_info=False, ufo_default=[], glyphs_multivalued=True
    )
)



"""Propagating anchors from components to their composites

Glyphs.app has a nice feature where anchors defined in the components
of composite glyphs are copied into the composites themselves. This feature
however is not very extensively documented.
The code here is based off the Rust implementation in fontc:

  https://github.com/googlefonts/fontc/blob/85795bf/glyphs-reader/src/propagate_anchors.rs

The latter is in turn based off the original Objective-C implementation, which was
shared with us privately.
"""

from __future__ import annotations

import logging
from collections import deque
from itertools import chain
from math import atan2, degrees, isinf
from typing import TYPE_CHECKING

from fontTools.misc.transform import Transform
from fontTools.ttLib.tables._g_l_y_f import GlyphCoordinates
from fontTools.varLib.models import VariationModel, normalizeLocation

from glyphsLib import glyphdata
from glyphsLib.classes import GSAnchor
from glyphsLib.types import Point


logger = logging.getLogger(__name__)


if TYPE_CHECKING:
    from typing import Iterable
    from glyphsLib.classes import GSComponent, GSFont, GSGlyph, GSLayer


def propagate_all_anchors(
    font: GSFont, *, glyph_data: glyphdata.GlyphData | None = None
) -> None:
    """Copy anchors from component glyphs into their including composites.

    If a custom `glyph_data` is provided, it will be used to override the
    category and subCategory of glyphs.
    """
    glyphs = {glyph.name: glyph for glyph in font.glyphs}
    propagate_all_anchors_impl(glyphs, font=font, glyph_data=glyph_data)


# the actual implementation, easier to test and compare with the original Rust code
def propagate_all_anchors_impl(
    glyphs: dict[str, GSGlyph],
    *,
    font: GSFont | None = None,
    glyph_data: glyphdata.GlyphData | None = None,
) -> None:
    # the reference implementation does this recursively, but we opt to
    # implement it by pre-sorting the work to ensure we always process components
    # first.
    todo = depth_sorted_composite_glyphs(glyphs)
    num_base_glyphs: dict[(str, str), int] = {}
    # NOTE: there's an important detail here, which is that we need to call the
    # 'anchors_traversing_components' function on each glyph, and save the returned
    # anchors, but we only *set* those anchors on glyphs that have components.
    # to make this work, we write the anchors to a separate data structure, and
    # then only update the actual glyphs after we've done all the work.
    all_anchors: dict[str, dict[str, list[GSAnchor]]] = {}

    # For brace layer interpolation
    master_locations, axes_triples = _get_design_space_info(font)
    layer_locations: dict[str, dict[str, float]] = {}
    variation_model_cache: dict = {}

    for name in todo:
        glyph = glyphs[name]
        for layer in _interesting_layers(glyph):
            # Record this layer's location before traversal so it's available
            # for interpolation of component anchors at brace layer locations
            loc = _get_layer_location(layer, master_locations)
            if loc is not None:
                layer_locations[layer.layerId] = loc
            anchors = anchors_traversing_components(
                glyph,
                layer,
                glyphs,
                all_anchors,
                num_base_glyphs,
                glyph_data,
                layer_locations=layer_locations,
                axes_triples=axes_triples,
                variation_model_cache=variation_model_cache,
            )
            maybe_log_new_anchors(anchors, glyph, layer)
            all_anchors.setdefault(name, {})[layer.layerId] = anchors

    # finally update our glyphs with the new anchors, where appropriate
    for name, layers in all_anchors.items():
        glyph = glyphs[name]
        if _has_components(glyph):
            for layer_id, layer_anchors in layers.items():
                glyph.layers[layer_id].anchors = layer_anchors


def maybe_log_new_anchors(
    anchors: list[GSAnchor], glyph: GSGlyph, layer: GSLayer
) -> None:
    if not _has_components(glyph) or not logger.isEnabledFor(logging.DEBUG):
        return
    prev_names = [a.name for a in layer.anchors]
    new_names = [a.name for a in anchors]
    if prev_names != new_names:
        logger.debug(
            "propagated anchors for ('%s': %s -> %s)",
            glyph.name,
            prev_names,
            new_names,
        )


def _is_master_layer(layer: GSLayer) -> bool:
    # Treat smart component layers as master layers
    return layer._is_master_layer or (
        layer.parent.smartComponentAxes and layer.smartComponentPoleMapping
    )


def _interesting_layers(glyph):
    return (
        l
        for l in glyph.layers
        if _is_master_layer(l) or l._is_bracket_layer() or l._is_brace_layer()
    )


def _has_components(glyph: GSGlyph) -> bool:
    return any(layer.components for layer in _interesting_layers(glyph))


def _get_category(
    glyph: GSGlyph,
    glyph_data: glyphdata.GlyphData | None = None,
) -> str:
    return (
        glyph.category
        or glyphdata.get_glyph(
            glyph.name, data=glyph_data, unicodes=glyph.unicodes
        ).category
    )


def _get_subCategory(
    glyph: GSGlyph,
    glyph_data: glyphdata.GlyphData | None = None,
) -> str:
    return (
        glyph.subCategory
        or glyphdata.get_glyph(
            glyph.name, data=glyph_data, unicodes=glyph.unicodes
        ).subCategory
    )


def _get_design_space_info(
    font: GSFont | None,
) -> tuple[dict[str, dict[str, float]], dict[str, tuple[float, float, float]]]:
    """Compute master locations and axis (min, default, max) triples from a font.

    Returns a (master_locations, axes_triples) tuple where:
    - master_locations maps master ID -> {axis_name: design_loc}
    - axes_triples maps axis_name -> (minimum, default, maximum)

    Returns empty dicts if font is None or has no masters.
    """
    if not font or not font.masters:
        return {}, {}

    from glyphsLib.builder.axes import get_axis_definitions, get_regular_master

    axis_defs = get_axis_definitions(font)
    default_master = get_regular_master(font)
    master_locations = {}
    for master in font.masters:
        master_locations[master.id] = {
            a.name: a.get_design_loc(master) for a in axis_defs
        }
    default_loc = master_locations[default_master.id]
    axes_triples = {}
    for axis_def in axis_defs:
        vals = [loc[axis_def.name] for loc in master_locations.values()]
        axes_triples[axis_def.name] = (min(vals), default_loc[axis_def.name], max(vals))
    return master_locations, axes_triples


def _get_layer_location(
    layer: GSLayer,
    master_locations: dict[str, dict[str, float]],
) -> dict[str, float] | None:
    """Return the design-space coordinates for a layer as a dict.

    For master layers, this is the master's axis values.
    For brace (intermediate) layers, this is the brace coordinates, filled
    with the associated master's values for any missing trailing axes.
    Returns None for other layer types (bracket, backup, etc.).
    """
    if not master_locations:
        return None
    if layer._is_master_layer:
        return master_locations[layer.layerId]
    if layer._is_brace_layer():
        master_loc = master_locations.get(layer.associatedMasterId)
        if master_loc is None:
            return None
        axis_names = list(master_loc.keys())
        coords = layer._brace_coordinates()
        # Fill missing trailing axes with the associated master's values
        loc = {}
        for i, name in enumerate(axis_names):
            if i < len(coords):
                loc[name] = coords[i]
            else:
                loc[name] = master_loc[name]
        return loc
    return None


def _interpolate_component_anchors(
    component_name: str,
    target_location: dict[str, float],
    all_anchors: dict[str, dict[str, list[GSAnchor]]],
    layer_locations: dict[str, dict[str, float]],
    axes_triples: dict[str, tuple[float, float, float]],
    variation_model_cache: dict,
) -> list[GSAnchor] | None:
    """Interpolate a component's anchors at a location where it has no source.

    Collects all available (location, anchors) pairs for the component, builds
    a VariationModel from normalized locations, and interpolates each anchor
    independently. Models are cached by source location set.

    Returns None if the component has no entries in all_anchors or interpolation
    fails entirely.
    """
    comp_layers = all_anchors.get(component_name)
    if not comp_layers:
        return None

    # Collect (normalized_location, anchors) pairs from master and brace layers.
    # Bracket layers are excluded: they define alternate glyph shapes whose
    # anchors shouldn't be interpolation sources for the default shape.
    # TODO (#1050): bracket+brace layers could be sources for alternate shapes.
    per_location: list[tuple[dict[str, float], list[GSAnchor]]] = []
    for layer_id, layer_anchors in comp_layers.items():
        loc = layer_locations.get(layer_id)
        if loc is None:
            continue
        norm_loc = normalizeLocation(loc, axes_triples)
        per_location.append((norm_loc, layer_anchors))

    if not per_location:
        return None

    norm_target = normalizeLocation(target_location, axes_triples)

    # Get canonical anchor names from the default source location.
    default_source_anchors = next(
        (
            anchors
            for norm_loc, anchors in per_location
            if all(v == 0.0 for v in norm_loc.values())
        ),
        None,
    )
    if default_source_anchors is None:
        logger.warning(
            "component '%s' has no default source for anchor interpolation; "
            "falling back to first available source",
            component_name,
        )
        default_source_anchors = per_location[0][1]
    anchor_names = [a.name for a in default_source_anchors]
    if not anchor_names:
        return []

    # Build per-anchor {normalized_location: position} maps
    per_anchor: dict[str, list[tuple[dict[str, float], Point]]] = {}
    for norm_loc, anchors in per_location:
        for anchor in anchors:
            if anchor.name in anchor_names:
                per_anchor.setdefault(anchor.name, []).append(
                    (norm_loc, anchor.position)
                )

    axis_order = list(axes_triples.keys())

    # Interpolate each anchor independently from its own set of source locations
    result = []
    for name in anchor_names:
        sources = per_anchor.get(name)
        if not sources:
            continue

        source_locs = [loc for loc, _ in sources]
        cache_key = frozenset(tuple(sorted(loc.items())) for loc in source_locs)
        if cache_key not in variation_model_cache:
            try:
                variation_model_cache[cache_key] = VariationModel(
                    source_locs, axisOrder=axis_order
                )
            except Exception as e:
                logger.warning(
                    "failed to build VariationModel for anchor '%s' on "
                    "component '%s': %s",
                    name,
                    component_name,
                    e,
                )
                continue

        model = variation_model_cache[cache_key]
        master_values = [GlyphCoordinates([(pos.x, pos.y)]) for _, pos in sources]

        try:
            interpolated = model.interpolateFromMasters(norm_target, master_values)
        except Exception as e:
            logger.warning(
                "failed to interpolate anchor '%s' on component '%s': %s",
                name,
                component_name,
                e,
            )
            continue

        if interpolated:
            x, y = interpolated[0]
            result.append(GSAnchor(name=name, position=Point(x, y)))

    return result if result else None


def _interpolate_smart_component_anchors(
    layer: GSLayer,
    component: GSComponent,
    glyphs: dict[str, GSGlyph],
    done_anchors: dict[str, dict[str, list[GSAnchor]]],
    anchors: list[GSAnchor],
) -> None:
    from ..smart_components import get_smart_component_variation_model

    model, location, masters = get_smart_component_variation_model(layer, component)
    if model is not None:
        coords = [
            GlyphCoordinates(
                [
                    anchor.position
                    for anchor in get_component_layer_anchors(
                        component, master, glyphs, done_anchors
                    )
                ]
            )
            for master in masters
        ]

        try:
            new_coords = model.interpolateFromMasters(location, coords)
        except Exception as e:
            raise ValueError(
                "Could not interpolate smart component %s used in %s"
                % (component.name, layer)
            ) from e

        for anchor, new_coord in zip(anchors, new_coords):
            anchor.position = Point(new_coord[0], new_coord[1])


def _get_base_glyph_count(
    base_glyph_counts: dict[(str, str), int],
    component_name: str,
    layer: GSLayer,
) -> int:
    """Look up the base glyph count for a component, handling bracket layers.

    Synthesized bracket layers on composites have different layerIds than their
    component's bracket layers. When the exact (name, layerId) lookup misses,
    fall back to the associated master's layer.
    """
    count = base_glyph_counts.get((component_name, layer.layerId))
    if count is not None:
        return count
    return base_glyph_counts.get((component_name, layer.associatedMasterId), 0)


def _resolve_component_anchors(
    component: GSComponent,
    layer: GSLayer,
    glyph: GSGlyph,
    glyphs: dict[str, GSGlyph],
    done_anchors: dict[str, dict[str, list[GSAnchor]]],
    layer_locations: dict[str, dict[str, float]] | None,
    axes_triples: dict[str, tuple[float, float, float]] | None,
    variation_model_cache: dict | None,
) -> list[GSAnchor] | None:
    """Get anchors for a component, falling back to interpolation for brace layers."""
    anchors = get_component_layer_anchors(component, layer, glyphs, done_anchors)
    if anchors is not None:
        return anchors

    # Component doesn't have an explicit source at this location (e.g. the
    # composite has a brace layer that its component doesn't). Try to
    # interpolate anchors from the component's available sources.
    if (
        layer._is_brace_layer()
        and layer_locations is not None
        and axes_triples
        and variation_model_cache is not None
    ):
        target_loc = layer_locations.get(layer.layerId)
        if target_loc is None:
            logger.warning(
                "brace layer '%s' of glyph '%s' has no known "
                "design-space location; skipping anchor interpolation",
                layer.name,
                glyph.name,
            )
        else:
            anchors = _interpolate_component_anchors(
                component.name,
                target_loc,
                done_anchors,
                layer_locations,
                axes_triples,
                variation_model_cache,
            )

    return anchors


def anchors_traversing_components(
    glyph: GSGlyph,
    layer: GSLayer,
    glyphs: dict[str, GSGlyph],
    done_anchors: dict[str, dict[str, list[GSAnchor]]],
    base_glyph_counts: dict[(str, str), int],
    glyph_data: glyphdata.GlyphData | None = None,
    layer_locations: dict[str, dict[str, float]] | None = None,
    axes_triples: dict[str, tuple[float, float, float]] | None = None,
    variation_model_cache: dict | None = None,
) -> list[GSAnchor]:
    """Return the anchors for this glyph, including anchors from components

    This function is a reimplmentation of a similarly named function in glyphs.app.

    The logic for copying anchors from components into their containing composites
    is tricky. Anchors need to be adjusted in various ways:

    - a special "*origin" anchor may exist, which modifies the position of other anchors
    - if a component is flipped on the x or y axes, we rename "top" to "bottom"
      and/or "left" to "right"
    - we need to apply the transform from the component
    - we may need to rename an anchor when the component is part of a ligature glyph
    """
    if not layer.anchors and not layer.components:
        return []

    # if this is a mark and it has anchors, just return them
    # (as in, don't even look at the components)
    if layer.anchors and _get_category(glyph, glyph_data) == "Mark":
        return list(origin_adjusted_anchors(layer.anchors))

    is_ligature = _get_subCategory(glyph, glyph_data) == "Ligature"
    has_underscore = any(a.name.startswith("_") for a in layer.anchors)

    number_of_base_glyphs = 0
    all_anchors = {}

    for component_idx, component in enumerate(layer.components):
        # because we process dependencies first we know that all components
        # referenced have already been propagated
        anchors = _resolve_component_anchors(
            component,
            layer,
            glyph,
            glyphs,
            done_anchors,
            layer_locations,
            axes_triples,
            variation_model_cache,
        )
        if anchors is None:
            logger.debug(
                "could not get layer '%s' for component '%s' of glyph '%s'",
                layer.layerId,
                component.name,
                glyph.name,
            )
            continue

        if component.component and component.component.smartComponentAxes:
            # If this is a smart component, we need to interpolate the anchors
            _interpolate_smart_component_anchors(
                layer, component, glyphs, done_anchors, anchors
            )

        # if this component has an explicitly set attachment anchor, use it
        if component_idx > 0 and component.anchor:
            maybe_rename_component_anchor(component.anchor, anchors)

        component_number_of_base_glyphs = _get_base_glyph_count(
            base_glyph_counts, component.name, layer
        )

        comb_has_underscore = any(
            len(a.name) >= 2 and a.name.startswith("_") for a in anchors
        )
        comb_has_exit = any(a.name.startswith("exit") for a in anchors)
        if not (comb_has_underscore or comb_has_exit):
            # delete exit anchors we may have taken from earlier components
            # (since a glyph should only have one exit anchor, and logically its
            # at the end)
            all_anchors = {
                n: a for n, a in all_anchors.items() if not n.startswith("exit")
            }

        component_transform = Transform(*component.transform)
        xscale, yscale = get_xy_rotation(component_transform)
        for anchor in anchors:
            new_has_underscore = anchor.name.startswith("_")
            if (component_idx > 0 or has_underscore) and new_has_underscore:
                continue
            # skip entry anchors on non-first glyphs
            if component_idx > 0 and anchor.name.startswith("entry"):
                continue

            new_anchor_name = rename_anchor_for_scale(anchor.name, xscale, yscale)
            if (
                is_ligature
                and component_number_of_base_glyphs > 0
                and not new_has_underscore
                and not (
                    new_anchor_name.startswith("exit")
                    or new_anchor_name.startswith("entry")
                )
            ):
                # dealing with marks like top_1 on a ligature
                new_anchor_name = make_liga_anchor_name(
                    new_anchor_name, number_of_base_glyphs
                )

            apply_transform_to_anchor(anchor, component_transform)
            anchor.name = new_anchor_name
            all_anchors[anchor.name] = anchor
            has_underscore |= new_has_underscore

        number_of_base_glyphs += _get_base_glyph_count(
            base_glyph_counts, component.name, layer
        )

    # now we've handled all the anchors from components, so copy over anchors
    # that were explicitly defined on this layer:
    all_anchors.update((a.name, a) for a in origin_adjusted_anchors(layer.anchors))
    has_underscore_anchor = False
    has_mark_anchor = False
    component_count_from_anchors = 0

    # now we count how many components we have, based on our anchors
    for name in all_anchors.keys():
        has_underscore_anchor |= name.startswith("_")
        has_mark_anchor |= name[0].isalpha() and name[0].isascii() if name else False
        if (
            not is_ligature
            and number_of_base_glyphs == 0
            and not name.startswith("_")
            and not (name.startswith("exit") or name.startswith("entry"))
            and "_" in name
        ):
            suffix = name[name.index("_") + 1 :]
            # carets count space between components, so the last caret
            # is n_components - 1
            maybe_add_one = 1 if name.startswith("caret") else 0
            anchor_index = 0
            try:
                anchor_index = int(suffix) + maybe_add_one
            except ValueError:
                pass
            component_count_from_anchors = max(
                component_count_from_anchors, anchor_index
            )
    if not has_underscore_anchor and number_of_base_glyphs == 0 and has_mark_anchor:
        number_of_base_glyphs += 1
    number_of_base_glyphs = max(number_of_base_glyphs, component_count_from_anchors)

    if any(a.name == "_bottom" for a in layer.anchors):
        all_anchors.pop("top", None)
        all_anchors.pop("_top", None)
    if any(a.name == "_top" for a in layer.anchors):
        all_anchors.pop("bottom", None)
        all_anchors.pop("_bottom", None)

    base_glyph_counts[(glyph.name, layer.layerId)] = number_of_base_glyphs

    return list(all_anchors.values())


def origin_adjusted_anchors(anchors: list[GSAnchor]) -> Iterable[GSAnchor]:
    """Iterate over anchors taking into account the special "*origin" anchor

    If that anchor is present it will be used to adjust the positions of other
    anchors, and will not be included in the output.
    """
    origin = next((a.position for a in anchors if a.name == "*origin"), Point(0, 0))
    return (
        GSAnchor(
            name=a.name,
            position=Point(a.position.x - origin.x, a.position.y - origin.y),
            userData=dict(a.userData),
        )
        for a in anchors
        if a.name != "*origin"
    )


def get_xy_rotation(xform: Transform) -> tuple[float, float]:
    """Returns (x, y) where a negative value indicates axis is flipped"""
    # this is based on examining the behaviour of glyphs via the macro panel
    # and careful testing.
    a, b = xform[:2]
    # first take the rotation
    angle = atan2(b, a)
    # then remove the rotation, and take the scale
    rotated = xform.rotate(-angle)
    xscale, yscale = (rotated[0], rotated[3])
    # then invert the scale if the rotation was >= 180°
    if abs(degrees(angle) - 180) < 0.001:
        xscale = -xscale
        yscale = -yscale

    return xscale, yscale


def apply_transform_to_anchor(anchor: GSAnchor, transform: Transform) -> None:
    """Apply the transform but also do some rounding.

    So we don't have anchors with points like (512, 302.000000006).
    """
    x, y = anchor.position
    pos = transform.transformPoint((x, y))
    anchor.position = Point(round(pos[0], 6), round(pos[1], 6))


def maybe_rename_component_anchor(comp_name: str, anchors: list[GSAnchor]) -> None:
    # e.g, go from 'top' to 'top_1'
    if "_" not in comp_name:
        return
    sub_name = comp_name[: comp_name.index("_")]
    mark_name = f"_{sub_name}"
    if any(a.name == sub_name for a in anchors) and any(
        a.name == mark_name for a in anchors
    ):
        comp_anchor = next(a for a in anchors if a.name == sub_name)
        comp_anchor.name = comp_name


def make_liga_anchor_name(name: str, base_number: int) -> str:
    if "_" in name:
        # if this anchor already has a number (like 'top_2') we want to consider that
        name, suffix = name.split("_", 1)
        try:
            num = int(suffix)
        except ValueError:
            num = 1
        return f"{name}_{base_number + num}"
    # otherwise we're turning 'top' into 'top_N'
    return f"{name}_{base_number + 1}"


def rename_anchor_for_scale(name: str, xscale: float, yscale: float) -> str:
    """If a component is rotated, flip bottom/top, left/right, entry/exit"""

    def swap_pair(s: str, one: str, two: str) -> str:
        if one in s:
            return s.replace(one, two)
        elif two in s:
            return s.replace(two, one)
        return s

    if xscale < 0.0:
        name = swap_pair(name, "left", "right")
        name = swap_pair(name, "exit", "entry")
    if yscale < 0.0:
        name = swap_pair(name, "bottom", "top")

    return name


def get_component_layer_anchors(
    component: GSComponent,
    layer: GSLayer,
    glyphs: dict[str, GSGlyph],
    anchors: dict[str, dict[str, list[GSAnchor]]],
) -> list[GSAnchor] | None:
    if component.name not in anchors:
        return None  # nothing to propagate

    glyph = glyphs.get(component.name)
    if glyph is None:
        return None  # invalid component reference, skip

    layer_anchors = None

    parent_is_master = _is_master_layer(layer)
    parent_is_bracket = layer._is_bracket_layer()
    parent_is_brace = layer._is_brace_layer()
    parent_axis_rules = (
        [] if not parent_is_bracket else list(layer._bracket_axis_rules())
    )

    # Try matching: master layer (by layerId), bracket layer (by axis rules +
    # associated master), or brace layer (by coordinates).
    # TODO (#1050, #1018): A layer can be both bracket and brace; matching
    # should check both axis rules AND coordinates for such layers.
    # Using GSLayer.layerKey() (#1018) would handle all combinations.
    for comp_layer in _interesting_layers(glyph):
        if (
            parent_is_bracket
            and comp_layer._is_bracket_layer()
            and comp_layer.associatedMasterId == layer.associatedMasterId
            and (list(comp_layer._bracket_axis_rules()) == parent_axis_rules)
        ) or (parent_is_master and comp_layer.layerId == layer.layerId):
            layer_anchors = anchors[component.name][comp_layer.layerId]
            break
        if (
            parent_is_brace
            and comp_layer._is_brace_layer()
            and comp_layer._brace_coordinates() == layer._brace_coordinates()
            and comp_layer.layerId in anchors.get(component.name, {})
        ):
            layer_anchors = anchors[component.name][comp_layer.layerId]
            break

    # For brace layers, return None when no match is found so the caller can
    # try interpolation. For other layer types, fall back to the associated master.
    if layer_anchors is None and not parent_is_brace:
        layer_anchors = anchors[component.name][layer.associatedMasterId]

    if layer_anchors is not None:
        # return a copy as they may be modified in place
        layer_anchors = [
            GSAnchor(
                name=a.name,
                position=Point(a.position.x, a.position.y),
                userData=dict(a.userData),
            )
            for a in layer_anchors
        ]
    return layer_anchors


def compute_max_component_depths(glyphs: dict[str, GSGlyph]) -> dict[str, float]:
    queue = deque()
    # Returns a map of the maximum component depth of each glyph.
    # - a glyph with no components has depth 0,
    # - a glyph with a component has depth 1,
    # - a glyph with a component that itself has a component has depth 2, etc
    # - a glyph with a cyclical component reference has infinite depth, which is
    #   technically a source error
    depths = {}

    # for cycle detection; anytime a glyph is waiting for components (and so is
    # pushed to the back of the queue) we record its name and the length of the queue.
    # If we process the same glyph twice without the queue having gotten smaller
    # (meaning we have gone through everything in the queue) that means we aren't
    # making progress, and have a cycle.
    waiting_for_components = {}

    for name, glyph in glyphs.items():
        if _has_components(glyph):
            queue.append(glyph)
        else:
            depths[name] = 0

    while queue:
        next_glyph = queue.popleft()
        comp_names = {
            comp.name
            for comp in chain.from_iterable(
                l.components for l in _interesting_layers(next_glyph)
            )
            if comp.name in glyphs  # ignore missing components
        }
        if all(comp in depths for comp in comp_names):
            depths[next_glyph.name] = (
                max((depths[c] for c in comp_names), default=-1) + 1
            )
            waiting_for_components.pop(next_glyph.name, None)
        else:
            # else push to the back to try again after we've done the rest
            # (including the currently missing components)
            last_queue_len = waiting_for_components.get(next_glyph.name)
            waiting_for_components[next_glyph.name] = len(queue)
            if last_queue_len != len(queue):
                logger.debug("glyph '%s' is waiting for components", next_glyph.name)
                queue.append(next_glyph)
            else:
                depths[next_glyph.name] = float("inf")
                waiting_for_components.pop(next_glyph.name, None)
                logger.warning("glyph '%s' has cyclical components", next_glyph.name)

    assert not waiting_for_components
    assert len(depths) == len(glyphs)

    return depths


def depth_sorted_composite_glyphs(glyphs: dict[str, GSGlyph]) -> list[str]:
    depths = compute_max_component_depths(glyphs)
    # skip glyphs with infinite depth (cyclic dependencies)
    by_depth = sorted(
        (depth, name) for name, depth in depths.items() if not isinf(depth)
    )
    return [name for _, name in by_depth]



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import datetime
from glyphsLib.types import parse_datetime

UFO_FORMAT = "%Y/%m/%d %H:%M:%S"


def to_ufo_time(datetime_obj):
    """Format a datetime object as specified for UFOs."""
    return datetime_obj.strftime(UFO_FORMAT)


def from_ufo_time(string):
    """Parses a datetime as specified for UFOs into a datetime object."""
    return datetime.datetime.strptime(string, UFO_FORMAT)


def from_loose_ufo_time(string):
    """Parses a datetime as specified for UFOs into a datetime object,
    or as the Glyphs formet."""
    try:
        return from_ufo_time(string)
    except ValueError:
        return parse_datetime(string)


def to_ufo_color(color):
    if isinstance(color, str):
        color = [int(v) for v in color.split(",")]
    if len(color) == 2:
        # Greyscale color
        color = (color[0], color[0], color[0], color[1])
    elif len(color) == 5:
        # CMYK color
        raise NotImplementedError("CMYK colors are not supported")
    return tuple(c / 255 for c in color)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .constants import (
    UFO2FT_COLOR_LAYER_MAPPING_KEY,
    LAYER_ID_KEY,
    LAYER_ORDER_PREFIX,
    LAYER_ORDER_TEMP_USER_DATA_KEY,
)


def to_ufo_color_layer_names(self, master, ufo):
    for glyph in ufo.layers.defaultLayer:
        if UFO2FT_COLOR_LAYER_MAPPING_KEY in glyph.lib:
            glyph.lib[UFO2FT_COLOR_LAYER_MAPPING_KEY] = [
                (self._layer_map[v[0]], v[1])
                for v in glyph.lib[UFO2FT_COLOR_LAYER_MAPPING_KEY]
            ]


def to_ufo_layer(self, glyph, layer):
    ufo_font = self._sources[layer.associatedMasterId or layer.layerId].font

    layer_name = layer.name
    # Give color layers better names
    if layer._is_color_palette_layer():
        layer_name = f"color.{layer._color_palette_index()}"
    elif layer._is_brace_layer():
        layer_name = layer._brace_layer_name()

    if layer.associatedMasterId == layer.layerId:
        ufo_layer = ufo_font.layers.defaultLayer
    elif layer_name not in ufo_font.layers:
        ufo_layer = ufo_font.newLayer(layer_name)
    elif layer_name in ufo_font.layers and glyph.name in ufo_font.layers[layer_name]:
        self.logger.warning(
            "%s %s: Glyph %s, layer %s: Duplicate glyph layer name",
            ufo_font.info.familyName,
            ufo_font.info.styleName,
            glyph.name,
            layer_name,
        )
        n = 1
        new_layer_name = layer_name
        while new_layer_name in ufo_font.layers:
            new_layer_name = layer_name + " #" + repr(n)
            n += 1
        ufo_layer = ufo_font.newLayer(new_layer_name)
    else:
        ufo_layer = ufo_font.layers[layer_name]
    if self.minimize_glyphs_diffs:
        ufo_layer.lib[LAYER_ID_KEY] = layer.layerId
        ufo_layer.lib[LAYER_ORDER_PREFIX + glyph.name] = _layer_order_in_glyph(
            self, layer
        )
    self._layer_map[layer.layerId] = ufo_layer.name
    return ufo_layer


def to_ufo_background_layer(self, layer):
    ufo_font = self._sources[layer.associatedMasterId or layer.layerId].font
    if layer.associatedMasterId == layer.layerId:
        layer_name = "public.background"
    else:
        layer_name = layer.name + ".background"
    if layer_name not in ufo_font.layers:
        background_layer = ufo_font.newLayer(layer_name)
    else:
        background_layer = ufo_font.layers[layer_name]
    return background_layer


def _layer_order_in_glyph(self, layer):
    # TODO: optimize?
    for order, glyph_layer in enumerate(layer.parent.layers.values()):
        if glyph_layer is layer:
            return order
    return None


def to_glyphs_layer(self, ufo_layer, glyph, master):
    if ufo_layer is self._sources[master.id].font.layers.defaultLayer:
        layer = _get_or_make_foreground(self, glyph, master)
    elif ufo_layer.name == "public.background":
        master_layer = _get_or_make_foreground(self, glyph, master)
        layer = master_layer.background
    elif ufo_layer.name.endswith(".background"):
        # Find or create the foreground layer
        # TODO: (jany) add lib attribute to find foreground by layer id
        foreground_name = ufo_layer.name[: -len(".background")]
        foreground = next(
            (
                l
                for l in glyph.layers
                if l.name == foreground_name and l.associatedMasterId == master.id
            ),
            None,
        )
        if foreground is None:
            foreground = self.glyphs_module.GSLayer()
            foreground.name = foreground_name
            foreground.associatedMasterId = master.id
        layer = foreground.background
        # Background layers don't have an associated master id nor a name nor an id
    else:
        layer = next(
            (
                l
                for l in glyph.layers
                if l.name == ufo_layer.name and l.associatedMasterId == master.id
            ),
            None,
        )
        if layer is None:
            layer = self.glyphs_module.GSLayer()
        layer.associatedMasterId = master.id
        if LAYER_ID_KEY in ufo_layer.lib:
            layer.layerId = ufo_layer.lib[LAYER_ID_KEY]
        layer.name = ufo_layer.name
        glyph.layers.append(layer)
    order_key = LAYER_ORDER_PREFIX + glyph.name
    if order_key in ufo_layer.lib:
        order = ufo_layer.lib[order_key]
        layer.userData[LAYER_ORDER_TEMP_USER_DATA_KEY] = order
    return layer


def _get_or_make_foreground(self, glyph, master):
    layer = glyph.layers[master.id]
    if layer is None:
        layer = glyph.layers[master.id] = self.glyphs_module.GSLayer()
    layer.layerId = master.id
    layer.name = master.name
    return layer


def to_glyphs_layer_order(self, glyph):
    # TODO: (jany) ask for the rules of layer ordering inside a glyph
    # For now, order according to key in lib
    glyph.layers = sorted(glyph.layers, key=_layer_order)
    for layer in glyph.layers:
        if LAYER_ORDER_TEMP_USER_DATA_KEY in layer.userData:
            del layer.userData[LAYER_ORDER_TEMP_USER_DATA_KEY]


def _layer_order(layer):
    if LAYER_ORDER_TEMP_USER_DATA_KEY in layer.userData:
        return layer.userData[LAYER_ORDER_TEMP_USER_DATA_KEY]
    return float("inf")



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from glyphsLib import types
from glyphsLib.pens import _to_glyphs_node_type, _to_ufo_node_type


def to_ufo_paths(self, ufo_glyph, layer):
    """Draw .glyphs paths onto a pen."""
    pen = ufo_glyph.getPointPen()

    for path in layer.paths:
        # the list is changed below, otherwise you can't draw more than once
        # per session.
        nodes = list(path.nodes)

        pen.beginPath()

        if not nodes:
            pen.endPath()
            continue

        if not path.closed:
            node = nodes.pop(0)
            assert node.type == "line", "Open path starts with off-curve points"
            pen.addPoint(
                tuple(node.position),
                segmentType="move",
                name=node.userData.get("name"),
                identifier=node.userData.get("UFO.identifier"),
            )
        else:
            # In Glyphs.app, the starting node of a closed contour is always
            # stored at the end of the nodes list.
            nodes.insert(0, nodes.pop())

        for node in nodes:
            node_type = _to_ufo_node_type(node.type)
            pen.addPoint(
                tuple(node.position),
                segmentType=node_type,
                smooth=node.smooth,
                name=node.userData.get("name"),
                identifier=node.userData.get("UFO.identifier"),
            )
            # NOTE: Can't do path_index, node_index through enumeration here because we
            # might have changed node order.
            # A node's name will be stored as a UFO point's name attribute, so filter
            # it from the Glyph node user data to avoid storing duplicate information.
            node_user_data = {
                k: v
                for k, v in node.userData.items()
                if k not in ("UFO.identifier", "name")
            }
            self.to_ufo_node_user_data(ufo_glyph, node, node_user_data)
        pen.endPath()


def to_glyphs_paths(self, ufo_glyph, layer):
    # Keep track of path and node numbers, otherwise to_glyphs_node_user_data must
    # call _indices on every single GSNode, which is a huge performance drain.
    for contour_index, contour in enumerate(ufo_glyph):
        path = self.glyphs_module.GSPath()
        for point in contour:
            node = self.glyphs_module.GSNode()
            node.position = types.Point(point.x, point.y)
            node.type = _to_glyphs_node_type(point.segmentType)
            node.smooth = point.smooth
            node.name = point.name
            if point.identifier is not None:
                node.userData["UFO.identifier"] = point.identifier
            path.nodes.append(node)
        path.closed = not contour.open
        if not contour.open:
            path.nodes.append(path.nodes.pop(0))
        layer.paths.append(path)

        for node_index, node in enumerate(path.nodes):
            self.to_glyphs_node_user_data(ufo_glyph, node, contour_index, node_index)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import collections
import logging
import os

from fontTools.varLib.models import piecewiseLinearMap

import fontTools.designspaceLib
from glyphsLib.util import build_ufo_path

from .axes import (
    get_axis_definitions,
    get_regular_master,
    font_uses_axis_locations,
)
from .constants import UFO_FILENAME_CUSTOM_PARAM, UFO_FILENAME_KEY


logger = logging.getLogger(__name__)


def to_designspace_sources(self):
    regular_master = get_regular_master(self.font)
    for master in self.font.masters:
        _to_designspace_source(self, master, (master is regular_master))
    _to_designspace_source_layer(self)
    _warn_duplicate_master_locations(self)


def _warn_duplicate_master_locations(self):
    designspace = self._designspace
    master_locations = collections.defaultdict(list)
    for source in designspace.sources:
        master_locations[tuple(source.location.items())].append(source)
    duplicates = {l: s for l, s in master_locations.items() if len(s) > 1}
    if duplicates:
        msg = [
            "DesignSpace sources contain duplicate locations; "
            "varLib expects each master to define a unique location."
        ]
        if any(s.layerName for s in designspace.sources):
            msg.append(
                " Make sure that you used consistent 'brace layer' names"
                " in all the glyph layers that share the same location."
            )
        for loc, sources in sorted(duplicates.items()):
            msg.append(
                "\n  %r => %r"
                % ([s.layerName if s.layerName else s.name for s in sources], dict(loc))
            )
        logger.warning("".join(msg))


def _to_designspace_source(self, master, is_regular):
    source = self._sources[master.id]
    ufo = source.font

    if is_regular:
        source.copyLib = True
        source.copyInfo = True
        source.copyGroups = True
        source.copyFeatures = True

    source.familyName = ufo.info.familyName
    source.styleName = ufo.info.styleName
    # TODO: recover original source name from userData
    # UFO_SOURCE_NAME_KEY
    source.name = f"{source.familyName} {source.styleName}"

    if UFO_FILENAME_CUSTOM_PARAM in master.customParameters:
        source.filename = master.customParameters[UFO_FILENAME_CUSTOM_PARAM]
    elif UFO_FILENAME_KEY in master.userData:
        source.filename = master.userData[UFO_FILENAME_KEY]
    else:
        source.filename = build_ufo_path("", source.familyName, source.styleName)

    # Make sure UFO filenames are unique, lest we overwrite masters that
    # happen to have the same weight name. Careful, name clashes can also happen when
    # the masters have a UFO_FILENAME_CUSTOM_PARAM: when the user duplicates a master
    # in Glyphs but forgets to change the custom parameter.
    # Attention: The following is done regardless of where source.filename came from.
    # Depending on the duplicate's order in the master list, this could lead to the
    # legitimate master's filename getting an underscore appended!
    n = 1
    while any(
        s is not source and s.filename == source.filename
        for s in self._sources.values()
    ):
        filename_stem, filename_ext = os.path.splitext(source.filename)
        source.filename = f"{filename_stem}#{n}{filename_ext}"
        if (
            UFO_FILENAME_CUSTOM_PARAM in master.customParameters
            or UFO_FILENAME_KEY in master.userData
        ):
            logger.warning(
                "The master with the style name '{}' (ID {}) would be written to the "
                "same destination as another master. Change the "
                "master's custom parameter '{}' "
                "to select a different destination. Proceeding, but appending"
                " '#{}' to the filename on disk.".format(
                    source.styleName, master.id, UFO_FILENAME_CUSTOM_PARAM, n
                )
            )
        else:
            logger.warning(
                "The master with the style name '{}' (ID {}) has the same style name "
                "as another one. All masters should have distinctive "
                "(style) names. Use the 'Master name' custom parameter"
                " on a master to give it a unique name. Proceeding "
                "with an unchanged name, but appending '#{}' to the file"
                " name on disk.".format(source.styleName, master.id, n)
            )
        n += 1

    designspace_axis_tags = {a.tag for a in self.designspace.axes}
    location = {}
    for axis_def in get_axis_definitions(self.font):
        # Only write locations along defined axes
        if axis_def.tag in designspace_axis_tags:
            location[axis_def.name] = axis_def.get_design_loc(master)
    source.location = location


def _to_designspace_source_layer(self):
    # To construct a source layer, we need
    # 1. The Designspace source filename and font object which holds the layer.
    # 2. The (brace) layer name itself.
    # 3. The location of the intermediate master in the design space.
    # (For logging purposes, it's nice to know which glyphs contain the layer.)
    #
    # Note that a brace layer can be associated with different master layers (e.g. the
    # 'a' can have a '{400}' brace layer associated with 'Thin', and 'b''s can be
    # associte with 'Black').
    # Also note that if a brace layer name has less values than there are axes, they
    # are supposed to take on the values from the associated master as the missing
    # values.

    # First, collect all brace layers in the font and which glyphs and which masters
    # they belong to.
    layer_to_master_ids = collections.defaultdict(set)
    layer_to_glyph_names = collections.defaultdict(list)
    for glyph in self.font.glyphs:
        for layer in glyph.layers:
            if layer._is_brace_layer():
                key = (layer._brace_layer_name(), tuple(layer._brace_coordinates()))
                layer_to_master_ids[key].add(layer.associatedMasterId)
                layer_to_glyph_names[key].append(glyph.name)

    # Next, insert the brace layers in a defined location in the existing designspace.
    designspace = self._designspace
    layers_to_insert = collections.defaultdict(list)
    for key, master_ids in layer_to_master_ids.items():
        brace_coordinates = list(key[1])
        layer_name = key[0]
        for master_id in master_ids:
            # ... as they may need to be filled up with the values of the associated
            # master.
            master = self._sources[master_id]
            master_coordinates = brace_coordinates
            if len(master_coordinates) < len(designspace.axes):
                master_locations = [master.location[a.name] for a in designspace.axes]
                master_coordinates = (
                    brace_coordinates + master_locations[len(brace_coordinates) :]
                )
            elif len(master_coordinates) > len(designspace.axes):
                logger.warning(
                    "Glyph(s) %s, brace layer '%s' defines more locations than "
                    "there are design axes.",
                    layer_to_glyph_names[key],
                    layer_name,
                )

            # If we have more locations than axes, ignore the extra locations.
            layer_coordinates_mapping = collections.OrderedDict(
                (axis.name, location)
                for axis, location in zip(designspace.axes, master_coordinates)
            )

            s = fontTools.designspaceLib.SourceDescriptor()
            s.filename = master.filename
            s.font = master.font
            s.layerName = layer_name
            s.name = f"{master.name} {layer_name}"
            s.location = layer_coordinates_mapping

            # We collect all generated SourceDescriptors first, grouped by the masters
            # they belong to, so we can insert them in a defined order in the next step.
            layers_to_insert[master_id].append(s)

    # Splice brace layers into the appropriate location after their master.
    for master_id, brace_layers in layers_to_insert.items():
        master = self._sources[master_id]
        insert_index = designspace.sources.index(master) + 1
        brace_layers.sort(key=lambda x: tuple(x.location.values()))
        designspace.sources[insert_index:insert_index] = brace_layers


def to_glyphs_sources(self):
    for master in self.font.masters:
        _to_glyphs_source(self, master)


def _to_glyphs_source(self, master):
    source = self._sources[master.id]

    # Retrieve the master locations: weight, width, custom 0 - 1 - 2 - 3
    for axis_def in get_axis_definitions(self.font):
        try:
            design_location = source.location[axis_def.name]
        except KeyError:
            # The location does not have this axis?
            continue

        axis_def.set_design_loc(master, design_location)
        if font_uses_axis_locations(self.font):
            # The user location can be found by reading the mapping backwards
            mapping = []
            for axis in self.designspace.axes:
                if axis.tag == axis_def.tag:
                    mapping = axis.map
                    break
            reverse_mapping = {dl: ul for ul, dl in mapping}
            user_location = piecewiseLinearMap(design_location, reverse_mapping)
            axis_def.set_user_loc(master, user_location)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import OrderedDict
from pathlib import Path
import glyphsLib
import logging
import openstep_plist
import os
import sys


logger = logging.getLogger(__name__)


class Parser:
    """Parses Python dictionaries from Glyphs files."""

    def __init__(self, current_type=OrderedDict, format_version=2):
        self.current_type = current_type
        self.format_version = format_version

    def parse(self, d):
        try:
            if isinstance(d, str):
                d = self._fl7_format_clean(d)
                d = openstep_plist.loads(d, use_numbers=True)
            elif isinstance(d, bytes):
                d = self._fl7_format_clean(d)
                d = openstep_plist.loads(d.decode(), use_numbers=True)
            result = self._parse(d)
        except openstep_plist.parser.ParseError as e:
            raise ValueError("Failed to parse file") from e
        return result

    def _fl7_format_clean(self, d):
        """FontLab 7 glyphs source format exports include a final closing semicolon.
        This method removes the semicolon before passing the string to the parser."""
        # see https://github.com/googlefonts/fontmake/issues/806
        if isinstance(d, str):
            d = d.rstrip(";\n")
        elif isinstance(d, bytes):
            d = d.rstrip(b";\n")
        return d

    def _parse(self, d, new_type=None):
        self.current_type = new_type or self.current_type
        if isinstance(d, list):
            return self._parse_list(d, new_type)
        if isinstance(d, (dict, OrderedDict)):
            return self._parse_dict(d, new_type)
        return d

    def _parse_list(self, d, new_type=None):
        self.current_type = new_type or self.current_type
        return [self._parse(x, new_type) for x in d]

    def parse_into_object(self, res, value):
        return self._parse_dict_into_object(res, value)

    def _parse_dict(self, text, new_type=None):
        """Parse a dictionary from source text starting at i."""
        old_current_type = self.current_type
        new_type = new_type or self.current_type
        if new_type is None:
            # customparameter.value needs to be set from the found value
            new_type = dict
        elif type(new_type) == list:
            new_type = new_type[0]
        res = new_type()
        self._parse_dict_into_object(res, text)
        self.current_type = old_current_type
        return res

    def _parse_dict_into_object(self, res, d):
        for name in d.keys():
            sane_name = name.replace(".", "__")
            if hasattr(res, f"_parse_{sane_name}_dict"):
                getattr(res, f"_parse_{sane_name}_dict")(self, d[name])
            elif isinstance(res, (dict, OrderedDict)):
                result = self._parse(d[name])
                try:
                    res[name] = result
                except (TypeError, KeyError):  # hmmm...
                    res = {}  # ugly, this fixes nested dicts in customparameters
                    res[name] = result
            else:
                res[name] = d[name]


def load_glyphspackage(package_dir):
    package = Path(package_dir)
    infofile = package / "fontinfo.plist"
    orderfile = package / "order.plist"
    uistatefile = package / "UIState.plist"

    glyphorder = []
    if orderfile.exists():
        with open(orderfile, "r", encoding="utf-8") as order_fh:
            glyphorder = openstep_plist.load(order_fh)

    with open(infofile, "r", encoding="utf-8") as info_fh:
        data = openstep_plist.load(info_fh, use_numbers=True)

    if uistatefile.exists():
        with open(uistatefile, "r", encoding="utf-8") as uistate_fh:
            uistate = openstep_plist.load(uistate_fh, use_numbers=True)
            if "displayStrings" in uistate and "DisplayStrings" not in uistate:
                uistate["DisplayStrings"] = uistate.pop("displayStrings")
            data.update(uistate)

    data["glyphs"] = []
    for glyphfile in (package / "glyphs").glob("*.glyph"):
        with open(glyphfile, "r") as fh:
            glyph = openstep_plist.load(fh, use_numbers=True)
        data["glyphs"].append(glyph)
    # Sort according to glyphorder

    def sort_key(glyph):
        glyphname = glyph["glyphname"]
        if glyphname in glyphorder:
            return (0, glyphorder.index(glyphname))
        else:
            return (1, glyphname)

    data["glyphs"] = list(sorted(data["glyphs"], key=sort_key))

    return data


def load(file_or_path, font=None):
    """Read a .glyphs file. 'file_or_path' should be a (readable) file
    object, a file name, or in the case of a .glyphspackage file, a
    directory name. 'font' is an existing object to parse into, or None.
    Return a 'font' or a GSFont object.
    """
    logger.info("Parsing .glyphs file")
    if font is None:
        font = glyphsLib.classes.GSFont()
    p = Parser(current_type=font.__class__)
    if hasattr(file_or_path, "read"):
        data = openstep_plist.load(file_or_path, use_numbers=True)
    elif os.path.isdir(file_or_path):
        data = load_glyphspackage(file_or_path)
    else:
        fp = open(file_or_path, "r", encoding="utf-8")
        data = openstep_plist.load(fp, use_numbers=True)
    p.parse_into_object(font, data)
    return font


def loads(s):
    """Read a .glyphs file from a (unicode) str object, or from
    a UTF-8 encoded bytes object.
    Return a GSFont object.
    """
    p = Parser(current_type=glyphsLib.classes.GSFont)
    logger.info("Parsing .glyphs file")
    res = glyphsLib.classes.GSFont()
    p.parse_into_object(res, openstep_plist.loads(s, use_numbers=True))
    return res


def main(args=None):
    """Roundtrip the .glyphs file given as an argument."""
    for arg in args:
        glyphsLib.dump(load(arg), sys.stdout)


if __name__ == "__main__":
    main(sys.argv[1:])



from __future__ import annotations

import logging
import math
import os.path
import uuid

import pytest
from copy import deepcopy
from datetime import datetime
from typing import TYPE_CHECKING

from fontTools.misc.transform import Transform as Affine

from glyphsLib.classes import (
    GSAnchor,
    GSFont,
    GSFontMaster,
    GSGlyph,
    GSLayer,
    GSComponent,
    GSSmartComponentAxis,
)
from glyphsLib.glyphdata import get_glyph
from glyphsLib.types import Point, Transform
from glyphsLib.writer import dumps

from glyphsLib.builder.transformations.align_alternate_layers import (
    align_alternate_layers,
)
from glyphsLib.builder.transformations.propagate_anchors import (
    _get_design_space_info,
    compute_max_component_depths,
    get_xy_rotation,
    propagate_all_anchors,
    propagate_all_anchors_impl,
    depth_sorted_composite_glyphs,
)

if TYPE_CHECKING:
    from typing import Callable, Self

DATA = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), "data")


# Basically all the tests below are translated from:
# https://github.com/googlefonts/fontc/blob/ecc727d/glyphs-reader/src/propagate_anchors.rs#L423-L959
# This is to ensure that the Python implementation behaves the same way as the Rust one.


class GlyphSetBuilder:
    glyphs: dict[str, GSGlyph]

    def __init__(self):
        self.font = GSFont()
        self.font.format_version = 3

    def build(self) -> dict[str, GSGlyph]:
        return {g.name: g for g in self.font.glyphs}

    def add_glyph(self, name: str, build_fn: Callable[["GlyphBuilder"], None]) -> Self:
        glyph = GlyphBuilder(name, self.font)
        build_fn(glyph)
        # this inserts the glyph in the font and sets the glyph.parent attribute to it;
        # GSLayer._is_bracket_layer()/_bracket_axis_rules() require this to check the
        # glyph.parent.format_version to determine where to look for the axis rules
        self.font.glyphs.append(glyph.build())
        return self


class GlyphBuilder:
    def __init__(self, name: str, font: GSFont = None):
        info = get_glyph(name)
        self.glyph = glyph = GSGlyph()
        glyph.name = name
        glyph.unicode = info.unicode
        glyph.category = info.category
        glyph.subCategory = info.subCategory
        glyph.parent = font
        self.num_masters = 0
        self.num_alternates = 0
        self.add_master_layer()

    def build(self) -> GSGlyph:
        return self.glyph

    def add_master_layer(self) -> Self:
        layer = GSLayer()
        layer.name = layer.layerId = layer.associatedMasterId = (
            f"master-{self.num_masters}"
        )
        self.num_masters += 1
        self.num_alternates = 0
        self.glyph.layers.append(layer)
        self.current_layer = layer
        return self

    def add_backup_layer(self, associated_master_idx=0):
        layer = GSLayer()
        layer.name = datetime.now().isoformat()
        layer.layerId = str(uuid.uuid4()).upper()
        master_layer = self.glyph.layers[associated_master_idx]
        layer.associatedMasterId = master_layer.layerId
        self.glyph.layers.append(layer)
        self.current_layer = layer
        return self

    def add_bracket_layer(self, axis_rules):
        assert self.current_layer._is_master_layer
        associated_master_id = self.current_layer.layerId
        master_layer = self.glyph.layers[associated_master_id]
        layer = GSLayer()
        layer.name = f"{master_layer.name} bracket-{self.num_alternates}"
        layer.layerId = str(uuid.uuid4()).upper()
        layer.associatedMasterId = associated_master_id
        layer.attributes["axisRules"] = axis_rules
        self.num_alternates += 1
        self.glyph.layers.append(layer)
        self.current_layer = layer
        return self

    def add_brace_layer(self, coordinates, associated_master_idx=None):
        if associated_master_idx is None:
            # Find the most recent master layer
            associated_master_id = None
            for l in reversed(list(self.glyph.layers)):
                if l._is_master_layer:
                    associated_master_id = l.layerId
                    break
            assert associated_master_id is not None
        else:
            master_layer = self.glyph.layers[associated_master_idx]
            associated_master_id = master_layer.layerId
        layer = GSLayer()
        layer.name = f"{{{', '.join(str(c) for c in coordinates)}}}"
        layer.layerId = str(uuid.uuid4()).upper()
        layer.associatedMasterId = associated_master_id
        layer.attributes["coordinates"] = coordinates
        self.glyph.layers.append(layer)
        self.current_layer = layer
        return self

    def set_category(self, category: str) -> Self:
        self.glyph.category = category
        return self

    def set_subCategory(self, subCategory: str) -> Self:
        self.glyph.subCategory = subCategory
        return self

    def add_smartComponentAxis(
        self, name: str, top_value: float, bottom_value: float
    ) -> Self:
        axis = GSSmartComponentAxis()
        axis.name = name
        axis.topValue = top_value
        axis.bottomValue = bottom_value
        self.glyph.smartComponentAxes.append(axis)
        return self

    def set_smartComponentPoleMapping(self, name: str, pole: int) -> Self:
        self.current_layer.smartComponentPoleMapping[name] = pole
        return self

    def add_component(
        self,
        name: str,
        pos: tuple[float, float],
        smart_component_values: dict[str, float] = None,
    ) -> Self:
        component = GSComponent(name, offset=pos)
        self.current_layer.components.append(component)
        if smart_component_values:
            component.smartComponentValues = smart_component_values
        return self

    def rotate_component(self, degrees: float) -> Self:
        # Set an explicit translate + rotation for the component
        component = self.current_layer.components[-1]
        component.transform = Transform(
            *Affine(*component.transform).rotate(math.radians(degrees))
        )
        return self

    def add_component_anchor(self, name: str) -> Self:
        # add an explicit anchor to the last added component
        component = self.current_layer.components[-1]
        component.anchor = name
        return self

    def add_anchor(
        self,
        name: str,
        pos: tuple[float, float],
        userData: dict | None = None,
    ) -> Self:
        anchor = GSAnchor(name, Point(*pos), userData=userData)
        self.current_layer.anchors.append(anchor)
        return self


def make_glyph(name: str, components: list[str]) -> GSGlyph:
    builder = GlyphBuilder(name)
    for comp in components:
        builder.add_component(comp, (0, 0))  # pos doesn't matter for this test
    return builder.build()


def test_components_by_depth():
    glyphs = {
        name: make_glyph(name, components)
        for name, components in [
            ("A", []),
            ("E", []),
            ("acutecomb", []),
            ("brevecomb", []),
            ("brevecomb_acutecomb", ["acutecomb", "brevecomb"]),
            ("AE", ["A", "E"]),
            ("Aacute", ["A", "acutecomb"]),
            ("Aacutebreve", ["A", "brevecomb_acutecomb"]),
            ("AEacutebreve", ["AE", "brevecomb_acutecomb"]),
            ("acute", ["acutecomb.case"]),
            ("acutecomb.case", ["acutecomb.alt"]),
            ("acutecomb.alt", ["acute"]),
            ("grave", ["grave"]),
        ]
    }

    assert compute_max_component_depths(glyphs) == {
        "A": 0,
        "E": 0,
        "acutecomb": 0,
        "brevecomb": 0,
        "brevecomb_acutecomb": 1,
        "AE": 1,
        "Aacute": 1,
        "Aacutebreve": 2,
        "AEacutebreve": 2,
        "acute": float("inf"),
        "acutecomb.case": float("inf"),
        "acutecomb.alt": float("inf"),
        "grave": float("inf"),
    }

    assert depth_sorted_composite_glyphs(glyphs) == [
        "A",
        "E",
        "acutecomb",
        "brevecomb",
        "AE",
        "Aacute",
        "brevecomb_acutecomb",
        "AEacutebreve",
        "Aacutebreve",
        # cyclical composites are skipped
    ]


def assert_equal_gsobjects(object1, object2):
    # glyphsLib.classes objects don't implement __eq__, so we resort to compare
    # their serialized forms... Ugly but works :(
    assert dumps(object1) == dumps(object2)


def assert_equal_glyphsets(glyphs1, glyphs2):
    assert len(glyphs1) == len(glyphs2)
    assert glyphs1.keys() == glyphs2.keys()
    for name in glyphs1:
        assert_equal_gsobjects(glyphs1[name], glyphs2[name])


def assert_anchors(actual, expected):
    assert len(actual) == len(expected)
    for a, e in zip(actual, expected):
        assert a.name == e[0]
        assert a.position == Point(*e[1])
        assert dict(a.userData) == (e[2] if len(e) > 2 else {})


def test_no_components_anchors_are_unchanged():
    # derived from the observed behaviour of glyphs 3.2.2 (3259)
    glyphs = (
        GlyphSetBuilder()
        .add_glyph(
            "A",
            lambda glyph: (
                glyph.add_anchor("bottom", (234, 0))
                .add_anchor("ogonek", (411, 0))
                .add_anchor("top", (234, 810))
            ),
        )
        .add_glyph(
            "acutecomb",
            lambda glyph: (
                glyph.add_anchor("_top", (0, 578)).add_anchor("top", (0, 810))
            ),
        )
        .build()
    )

    glyphs2 = deepcopy(glyphs)
    propagate_all_anchors_impl(glyphs2)
    # nothing should change here
    assert_equal_glyphsets(glyphs, glyphs2)


def test_basic_composite_anchor():
    # derived from the observed behaviour of glyphs 3.2.2 (3259)
    glyphs = (
        GlyphSetBuilder()
        .add_glyph(
            "A",
            lambda glyph: (
                glyph.add_anchor("bottom", (234, 0))
                .add_anchor("ogonek", (411, 0))
                .add_anchor("top", (234, 810))
            ),
        )
        .add_glyph(
            "acutecomb",
            lambda glyph: (
                glyph.add_anchor("_top", (0, 578)).add_anchor("top", (0, 810))
            ),
        )
        .add_glyph(
            "Aacute",
            lambda glyph: (
                glyph.add_component("A", (0, 0)).add_component("acutecomb", (234, 232))
            ),
        )
        .build()
    )
    propagate_all_anchors_impl(glyphs)

    new_glyph = glyphs["Aacute"]
    assert_anchors(
        new_glyph.layers[0].anchors,
        [
            ("bottom", (234, 0)),
            ("ogonek", (411, 0)),
            ("top", (234, 1042)),
        ],
    )


def test_propagate_ligature_anchors():
    # derived from the observed behaviour of glyphs 3.2.2 (3259)
    # this is based on the IJ glyph in Oswald (ExtraLight)
    glyphs = (
        GlyphSetBuilder()
        .add_glyph(
            "I",
            lambda glyph: (
                glyph.add_anchor("bottom", (103, 0))
                .add_anchor("ogonek", (103, 0))
                .add_anchor("top", (103, 810))
                .add_anchor("topleft", (20, 810))
            ),
        )
        .add_glyph(
            "J",
            lambda glyph: (
                glyph.add_anchor("bottom", (133, 0)).add_anchor("top", (163, 810))
            ),
        )
        .add_glyph(
            "IJ",
            lambda glyph: (
                glyph.set_subCategory("Ligature")
                .add_component("I", (0, 0))
                .add_component("J", (206, 0))
            ),
        )
        .build()
    )
    propagate_all_anchors_impl(glyphs)
    ij = glyphs["IJ"]
    # these were derived by running the built in glyphs.app propagate anchors
    # method from the macro panel
    assert_anchors(
        ij.layers[0].anchors,
        [
            ("bottom_1", (103, 0)),
            ("ogonek_1", (103, 0)),
            ("top_1", (103, 810)),
            ("topleft_1", (20, 810)),
            ("bottom_2", (339, 0)),
            ("top_2", (369, 810)),
        ],
    )


def test_digraphs_arent_ligatures():
    # derived from the observed behaviour of glyphs 3.2.2 (3259)
    # this is based on the IJ glyph in Oswald (ExtraLight)
    glyphs = (
        GlyphSetBuilder()
        .add_glyph(
            "I",
            lambda glyph: (
                glyph.add_anchor("bottom", (103, 0))
                .add_anchor("ogonek", (103, 0))
                .add_anchor("top", (103, 810))
                .add_anchor("topleft", (20, 810))
            ),
        )
        .add_glyph(
            "J",
            lambda glyph: (
                glyph.add_anchor("bottom", (133, 0)).add_anchor("top", (163, 810))
            ),
        )
        .add_glyph(
            "IJ",
            lambda glyph: (
                glyph.add_component("I", (0, 0)).add_component("J", (206, 0))
            ),
        )
        .build()
    )
    propagate_all_anchors_impl(glyphs)
    ij = glyphs["IJ"]
    # these were derived by running the built in glyphs.app propagate anchors
    # method from the macro panel
    assert_anchors(
        ij.layers[0].anchors,
        # 'J' component comes last; the 'bottom' and 'top' anchors are from 'J'
        # shifted by the 'J' component's offset (206, 0).
        # 'ogonek' and 'topleft' are inherited from 'I', the first component, which
        # has (0, 0) offset hence the same anchor positions as the original 'I' glyph.
        [
            ("bottom", (339, 0)),
            ("ogonek", (103, 0)),
            ("top", (369, 810)),
            ("topleft", (20, 810)),
        ],
    )


def test_propagate_across_layers(caplog):
    # derived from the observed behaviour of glyphs 3.4 (3414)
    glyphs = (
        GlyphSetBuilder()
        .add_glyph(
            "A",
            lambda glyph: (
                glyph.add_anchor("bottom", (206, 16))
                .add_anchor("ogonek", (360, 13))
                .add_anchor("top", (212, 724))
                .add_bracket_layer([{"min": 500}])
                .add_anchor("bottom", (206, 16))
                .add_anchor("top", (212, 724))
                .add_master_layer()
                .add_anchor("bottom", (278, 12))
                .add_anchor("ogonek", (464, 13))
                .add_anchor("top", (281, 758))
                .add_bracket_layer([{"min": 500}])
                .add_anchor("bottom", (278, 12))
                .add_anchor("top", (281, 758))
                .add_backup_layer()
                .add_anchor("top", (290, 690))
            ),
        )
        .add_glyph(
            "acutecomb",
            lambda glyph: (
                glyph.add_anchor("_top", (150, 580))
                .add_anchor("top", (170, 792))
                .add_master_layer()
                .add_anchor("_top", (167, 580))
                .add_anchor("top", (170, 792))
                .add_backup_layer()
                .add_anchor("_top", (335, 502))
            ),
        )
        .add_glyph(
            "Aacute",
            lambda glyph: (
                glyph.add_component("A", (0, 0))
                .add_component("acutecomb", (62, 144))
                .add_bracket_layer([{"min": 500}])
                .add_component("A", (20, 0))
                .add_component("acutecomb", (82, 144))
                .add_master_layer()
                .add_component("A", (0, 0))
                .add_component("acutecomb", (114, 178))
                .add_bracket_layer([{"min": 500}])
                .add_component("A", (30, 0))
                .add_component("acutecomb", (144, 178))
                .add_backup_layer()
                .add_component("A", (0, 0))
                .add_component("acutecomb", (-45, 188))
            ),
        )
        .add_glyph(
            "acutecomb.case",
            lambda glyph: (
                glyph.add_component("acutecomb", (0, 0))
                .add_master_layer()
                .add_component("acutecomb", (0, 0))
                # this backup layer contains a potentially cyclical reference
                # as the component's base glyph in turn points back at self;
                # this doesn't trigger an infinite loop because backup layers
                # are skipped when propagating anchors
                .add_backup_layer()
                .add_component("acute", (0, 0))
            ),
        )
        .add_glyph(
            "acute",
            lambda glyph: (
                glyph.add_component("acutecomb.case", (0, 0))
                .add_master_layer()
                .add_component("acutecomb.case", (0, 0))
            ),
        )
        .build()
    )
    with caplog.at_level(logging.WARNING):
        propagate_all_anchors_impl(glyphs)

    new_glyph = glyphs["Aacute"]
    assert_anchors(
        new_glyph.layers["master-0"].anchors,
        [
            ("bottom", (206, 16)),
            ("ogonek", (360, 13)),
            ("top", (232, 936)),
        ],
    )

    assert_anchors(
        new_glyph.layers["master-1"].anchors,
        [
            ("bottom", (278, 12)),
            ("ogonek", (464, 13)),
            ("top", (284, 970)),
        ],
    )

    # alternate 'bracket' layers should work as well
    alternate_layers = [l for l in new_glyph.layers if l._is_bracket_layer()]
    assert len(alternate_layers) == 2
    assert_anchors(
        alternate_layers[0].anchors,
        [
            ("bottom", (226, 16)),
            ("top", (252, 936)),
        ],
    )
    assert_anchors(
        alternate_layers[1].anchors,
        [
            ("bottom", (308, 12)),
            ("top", (314, 970)),
        ],
    )

    # non-master (e.g. backup) layers are silently skipped
    assert not new_glyph.layers[-1]._is_master_layer
    assert not new_glyph.layers[-1]._is_bracket_layer()
    assert_anchors(new_glyph.layers[-1].anchors, [])

    assert len(caplog.records) == 0


def test_propagate_across_layers_with_circular_reference(caplog):
    glyphs = (
        GlyphSetBuilder()
        # acutecomb.alt contains a cyclical reference to itself in its master layer
        # test that this doesn't cause an infinite loop
        .add_glyph(
            "acutecomb.alt",
            lambda glyph: (
                glyph.add_component("acutecomb.alt", (0, 0))
                .add_master_layer()
                .add_component("acutecomb.alt", (0, 0))
            ),
        )
        # gravecomb and grave contain cyclical component references to one another
        # in their master layers; test that this doesn't cause an infinite loop either
        .add_glyph(
            "gravecomb",
            lambda glyph: (
                glyph.add_component("grave", (0, 0))
                .add_master_layer()
                .add_component("grave", (0, 0))
            ),
        )
        .add_glyph(
            "grave",
            lambda glyph: (
                glyph.add_component("gravecomb", (0, 0))
                .add_master_layer()
                .add_component("gravecomb", (0, 0))
            ),
        )
        .build()
    )

    with caplog.at_level(logging.WARNING):
        propagate_all_anchors_impl(glyphs)

    assert len(caplog.records) == 2
    assert caplog.records[0].message == "glyph 'acutecomb.alt' has cyclical components"
    assert caplog.records[1].message == "glyph 'gravecomb' has cyclical components"


def test_remove_exit_anchor_on_component():
    # derived from the observed behaviour of glyphs 3.2.2 (3259)
    glyphs = (
        GlyphSetBuilder()
        .add_glyph("comma", lambda glyph: ())
        .add_glyph(
            "ain-ar.init",
            lambda glyph: (
                glyph.add_anchor("top", (294, 514)).add_anchor("exit", (0, 0))
            ),
        )
        .add_glyph(
            "ain-ar.init.alt",
            lambda glyph: (
                glyph.add_component("ain-ar.init", (0, 0)).add_component(
                    "comma", (0, 0)
                )
            ),
        )
        .build()
    )
    propagate_all_anchors_impl(glyphs)

    new_glyph = glyphs["ain-ar.init.alt"]
    assert_anchors(new_glyph.layers[0].anchors, [("top", (294, 514))])


def test_component_anchor():
    # derived from the observed behaviour of glyphs 3.2.2 (3259)
    glyphs = (
        GlyphSetBuilder()
        .add_glyph(
            "acutecomb",
            lambda glyph: (
                glyph.add_anchor("_top", (150, 580)).add_anchor("top", (170, 792))
            ),
        )
        .add_glyph(
            "aa",
            lambda glyph: (
                glyph.add_anchor("bottom_1", (218, 8))
                .add_anchor("bottom_2", (742, 7))
                .add_anchor("ogonek_1", (398, 9))
                .add_anchor("ogonek_2", (902, 9))
                .add_anchor("top_1", (227, 548))
                .add_anchor("top_2", (746, 548))
            ),
        )
        .add_glyph(
            "a_a",
            lambda glyph: glyph.add_component("aa", (0, 0)),
        )
        .add_glyph(
            "a_aacute",
            lambda glyph: (
                glyph.add_component("a_a", (0, 0))
                .add_component("acutecomb", (596, -32))
                .add_component_anchor("top_2")
            ),
        )
        .build()
    )
    propagate_all_anchors_impl(glyphs)

    new_glyph = glyphs["a_aacute"]
    assert_anchors(
        new_glyph.layers[0].anchors,
        [
            ("bottom_1", (218, 8)),
            ("bottom_2", (742, 7)),
            ("ogonek_1", (398, 9)),
            ("ogonek_2", (902, 9)),
            ("top_1", (227, 548)),
            ("top_2", (766, 760)),
        ],
    )


def test_origin_anchor():
    # derived from the observed behaviour of glyphs 3.2.2 (3259)
    glyphs = (
        GlyphSetBuilder()
        .add_glyph(
            "a",
            lambda glyph: (
                glyph.add_anchor("*origin", (-20, 0))
                .add_anchor("bottom", (242, 7))
                .add_anchor("ogonek", (402, 9))
                .add_anchor("top", (246, 548))
            ),
        )
        .add_glyph(
            "acutecomb",
            lambda glyph: (
                glyph.add_anchor("_top", (150, 580)).add_anchor("top", (170, 792))
            ),
        )
        .add_glyph(
            "aacute",
            lambda glyph: (
                glyph.add_component("a", (0, 0)).add_component("acutecomb", (116, -32))
            ),
        )
        .build()
    )
    propagate_all_anchors_impl(glyphs)

    new_glyph = glyphs["aacute"]
    assert_anchors(
        new_glyph.layers[0].anchors,
        [
            ("bottom", (262, 7)),
            ("ogonek", (422, 9)),
            ("top", (286, 760)),
        ],
    )


def test_contextual_anchors():
    glyphs = (
        GlyphSetBuilder()
        .add_glyph(
            "behDotless-ar.init",
            lambda glyph: (
                glyph.add_anchor("bottom", (50, 0))
                .add_anchor(
                    "*bottom",
                    (95, 0),
                    userData={"GPOS_Context": "* behDotless-ar.medi"},
                )
                .add_anchor("top", (35, 229))
            ),
        )
        .add_glyph(
            "behDotless-ar.medi",
            lambda glyph: (
                glyph.add_component("behDotless-ar.init", (0, 0)).add_anchor(
                    "*bottom",
                    (95, 0),
                    userData={"GPOS_Context": "* behDotless-ar.fina"},
                )
            ),
        )
        .add_glyph(
            "behDotless-ar.fina",
            lambda glyph: glyph.add_component("behDotless-ar.init", (0, 0)),
        )
        .build()
    )
    propagate_all_anchors_impl(glyphs)

    new_glyph = glyphs["behDotless-ar.medi"]
    assert_anchors(
        new_glyph.layers[0].anchors,
        [
            ("bottom", (50, 0)),
            ("*bottom", (95, 0), {"GPOS_Context": "* behDotless-ar.fina"}),
            ("top", (35, 229)),
        ],
    )

    new_glyph = glyphs["behDotless-ar.fina"]
    assert_anchors(
        new_glyph.layers[0].anchors,
        [
            ("bottom", (50, 0)),
            ("*bottom", (95, 0), {"GPOS_Context": "* behDotless-ar.medi"}),
            ("top", (35, 229)),
        ],
    )


def test_invert_names_on_rotation():
    # derived from the observed behaviour of glyphs 3.2.2 (3259)
    glyphs = (
        GlyphSetBuilder()
        .add_glyph("comma", lambda glyph: ())
        .add_glyph(
            "commaaccentcomb",
            lambda glyph: (
                glyph.add_anchor("_bottom", (289, 0))
                .add_anchor("mybottom", (277, -308))
                .add_component("comma", (9, -164))
            ),
        )
        .add_glyph(
            "commaturnedabovecomb",
            lambda glyph: (
                glyph.add_component("commaaccentcomb", (589, 502)).rotate_component(180)
            ),
        )
        .build()
    )
    propagate_all_anchors_impl(glyphs)

    new_glyph = glyphs["commaturnedabovecomb"]
    assert_anchors(
        new_glyph.layers[0].anchors,
        [("_top", (300, 502)), ("mytop", (312, 810))],
    )


def test_affine_scale():
    assert get_xy_rotation(Affine().translate(589, 502).rotate(math.radians(180))) == (
        -1,
        -1,
    )
    assert get_xy_rotation(Affine().translate(10, 10)) == (1, 1)
    assert get_xy_rotation(Affine().scale(1, -1)) == (1, -1)
    assert get_xy_rotation(Affine().scale(-1, 1)) == (-1, 1)
    assert get_xy_rotation(
        Affine().translate(589, 502).rotate(math.radians(180)).scale(-1, 1)
    ) == (
        1,
        -1,
    )


def test_real_files():
    # the tricky parts of these files have been factored out into separate tests,
    # but we'll keep them in case there are other regressions lurking
    expected = GSFont(os.path.join(DATA, "PropagateAnchorsTest-propagated.glyphs"))
    font = GSFont(os.path.join(DATA, "PropagateAnchorsTest.glyphs"))

    propagate_all_anchors(font)

    assert len(font.glyphs) == len(expected.glyphs)
    assert [g.name for g in font.glyphs] == [g.name for g in expected.glyphs]
    for g1, g2 in zip(font.glyphs, expected.glyphs):
        assert len(g1.layers) == len(g2.layers)
        for l1, l2 in zip(g1.layers, g2.layers):
            # seems like the latest Glyphs.app insists on sorting anchors alphabetically
            # whereas we keep the original order; but what matters is their name and
            # position
            assert sorted((a.name, tuple(a.position)) for a in l1.anchors) == sorted(
                (a.name, tuple(a.position)) for a in l2.anchors
            )


@pytest.mark.parametrize(
    "test_file",
    [
        "AlignAlternateLayers-g2.glyphs",
        "AlignAlternateLayers-g3.glyphs",
    ],
)
def test_propagate_anchors_after_aligining_alternates(test_file):
    # https://github.com/googlefonts/glyphsLib/issues/1090
    font = GSFont(os.path.join(DATA, test_file))

    Cacute = font.glyphs["Cacute"]
    assert not any(l._is_bracket_layer() for l in Cacute.layers)
    for layer in Cacute.layers:
        assert len(layer.anchors) == 0

    align_alternate_layers(font)
    propagate_all_anchors(font)

    # all layers have now 2 anchors, including the new alternates
    Cacute = font.glyphs["Cacute"]
    alternate_layers = [l for l in Cacute.layers if l._is_bracket_layer()]
    assert len(alternate_layers) == 3
    for layer in Cacute.layers:
        assert len(layer.anchors) == 2
        assert [a.name for a in layer.anchors] == ["bottom", "top"]

    # the 'bottom' anchor in one of the alternate layers has different
    # position than the corresponding anchor in the associated master layer
    # (i.e. it was propagated and not simply duplicated)
    alt_layer = alternate_layers[0]
    alt_bottom = alt_layer.anchors[0]
    master_layer = Cacute.layers[alt_layer.associatedMasterId]
    master_bottom = master_layer.anchors[0]
    assert alt_bottom.name == master_bottom.name == "bottom"
    assert tuple(alt_bottom.position) == (329, 0)
    assert tuple(master_bottom.position) == (301, 0)


def test_entry_anchor_on_non_first_component():
    glyphs = (
        GlyphSetBuilder()
        .add_glyph(
            "part1",
            lambda glyph: (glyph.add_anchor("top", (10, 0))),
        )
        .add_glyph(
            "part2",
            lambda glyph: (glyph.add_anchor("entry.2", (10, 0))),
        )
        .add_glyph(
            "combo",
            lambda glyph: (
                glyph.add_component("part1", (0, 0)).add_component("part2", (100, 0))
            ),
        )
        .build()
    )
    propagate_all_anchors_impl(glyphs)

    new_glyph = glyphs["combo"]
    assert_anchors(new_glyph.layers[0].anchors, [("top", (10, 0))])


def test_cursive_anchors_ligature():
    glyphs = (
        GlyphSetBuilder()
        .add_glyph(
            "part1_part2",
            lambda glyph: (
                glyph.add_anchor("entry.1", (10, 0)).add_anchor("exit.1", (100, 0))
            ),
        )
        .add_glyph(
            "combo",
            lambda glyph: (
                glyph.set_subCategory("Ligature").add_component("part1_part2", (0, 0))
            ),
        )
        .build()
    )
    propagate_all_anchors_impl(glyphs)

    new_glyph = glyphs["combo"]
    assert_anchors(
        new_glyph.layers[0].anchors,
        [
            ("entry.1", (10, 0)),
            ("exit.1", (100, 0)),
        ],
    )


def test_smart_component_anchors():
    glyphs = (
        GlyphSetBuilder()
        .add_glyph(
            "base",
            lambda glyph: (
                glyph.add_smartComponentAxis("TEST", 100, -100)
                .set_smartComponentPoleMapping("TEST", 1)
                .add_anchor("top", (23, 103))
                .add_anchor("bottom", (36, -51))
                .add_backup_layer()
                .set_smartComponentPoleMapping("TEST", 2)
                .add_anchor("top", (33, 123))
                .add_anchor("bottom", (36, -51))
            ),
        )
        .add_glyph(
            "smart1",
            lambda glyph: (glyph.add_component("base", (0, 0), {"TEST": -100})),
        )
        .add_glyph(
            "smart2",
            lambda glyph: (glyph.add_component("base", (0, 0), {"TEST": 100})),
        )
        .add_glyph(
            "smart3",
            lambda glyph: (glyph.add_component("base", (0, 0), {"TEST": 0})),
        )
        .add_glyph(
            "smart4",
            lambda glyph: (glyph.add_component("base", (20, 10), {"TEST": 0})),
        )
        .build()
    )
    propagate_all_anchors_impl(glyphs)

    assert_anchors(
        glyphs["smart1"].layers[0].anchors,
        [
            ("top", (23, 103)),
            ("bottom", (36, -51)),
        ],
    )
    assert_anchors(
        glyphs["smart2"].layers[0].anchors,
        [
            ("top", (33, 123)),
            ("bottom", (36, -51)),
        ],
    )
    assert_anchors(
        glyphs["smart3"].layers[0].anchors,
        [
            ("top", (28, 113)),
            ("bottom", (36, -51)),
        ],
    )
    assert_anchors(
        glyphs["smart4"].layers[0].anchors,
        [
            ("top", (48, 123)),
            ("bottom", (56, -41)),
        ],
    )


def test_bracket_ligature_anchor_numbering():
    """Ligature anchor numbering must work on synthesized bracket layers.

    When align_alternate_layers creates a bracket layer on a ligature composite
    (because its component has bracket layers), the synthesized layer has a
    different layerId than the component's bracket layer. The base_glyph_counts
    lookup must handle this mismatch, or ligature numbering is skipped.
    """
    axis_rules = [{"min": 160, "max": None}]

    builder = GlyphSetBuilder()
    # Add a master so align_alternate_layers can find master IDs
    master = GSFontMaster()
    master.id = "master-0"
    builder.font.masters.append(master)
    # "f" - simple base glyph with a top anchor
    builder.add_glyph(
        "f",
        lambda glyph: glyph.add_anchor("top", (100, 700)),
    )
    # "h" - base glyph with a bracket layer (different top anchor position)
    builder.add_glyph(
        "h",
        lambda glyph: (
            glyph.add_anchor("top", (200, 700))
            .add_bracket_layer(axis_rules)
            .add_anchor("top", (200, 600))
        ),
    )
    # "f_h" - ligature composite of f + h (no bracket layers of its own)
    builder.add_glyph(
        "f_h",
        lambda glyph: (
            glyph.set_subCategory("Ligature")
            .add_component("f", (0, 0))
            .add_component("h", (300, 0))
        ),
    )

    glyphs = builder.build()

    # align_alternate_layers synthesizes a bracket layer on f_h
    align_alternate_layers(builder.font)
    propagate_all_anchors_impl(glyphs)

    f_h = glyphs["f_h"]
    master_layers = [l for l in f_h.layers if l._is_master_layer]
    bracket_layers = [l for l in f_h.layers if l._is_bracket_layer()]

    assert len(bracket_layers) == 1

    # Master layer should have ligature-numbered anchors
    assert_anchors(
        master_layers[0].anchors,
        [("top_1", (100, 700)), ("top_2", (500, 700))],
    )
    # Bracket layer must also have ligature-numbered anchors (not plain "top")
    assert_anchors(
        bracket_layers[0].anchors,
        [("top_1", (100, 700)), ("top_2", (500, 600))],
    )


def test_get_design_space_info_none():
    """Returns empty dicts when font is None or has no masters."""
    assert _get_design_space_info(None) == ({}, {})
    assert _get_design_space_info(GSFont()) == ({}, {})


def test_get_design_space_info_single_axis():
    """Computes correct locations and triples for a single-axis font."""
    font = GSFont()
    font.format_version = 3
    master0 = GSFontMaster()
    master0.id = "light"
    master0.axes = [100]
    font.masters.append(master0)
    master1 = GSFontMaster()
    master1.id = "bold"
    master1.axes = [200]
    font.masters.append(master1)

    master_locations, axes_triples = _get_design_space_info(font)

    assert master_locations["light"]["Weight"] == 100
    assert master_locations["bold"]["Weight"] == 200
    # Default is first master (light), so triple is (min=100, default=100, max=200)
    assert axes_triples["Weight"] == (100, 100, 200)


def test_get_design_space_info_two_axes():
    """Computes correct locations and triples for a two-axis font."""
    font = GSFont()
    font.format_version = 3
    m0 = GSFontMaster()
    m0.id = "light-narrow"
    m0.axes = [100, 75]
    font.masters.append(m0)
    m1 = GSFontMaster()
    m1.id = "bold-narrow"
    m1.axes = [200, 75]
    font.masters.append(m1)
    m2 = GSFontMaster()
    m2.id = "bold-wide"
    m2.axes = [200, 100]
    font.masters.append(m2)

    master_locations, axes_triples = _get_design_space_info(font)

    assert len(master_locations) == 3
    assert master_locations["light-narrow"]["Weight"] == 100
    assert master_locations["light-narrow"]["Width"] == 75
    assert master_locations["bold-wide"]["Weight"] == 200
    assert master_locations["bold-wide"]["Width"] == 100
    # Default is first master
    assert axes_triples["Weight"] == (100, 100, 200)
    assert axes_triples["Width"] == (75, 75, 100)


def test_interpolate_brace_layer_component_anchors():
    """When a composite has a brace layer but its component doesn't, interpolate
    the component's anchor positions from its available master sources.

    See: https://github.com/googlefonts/fontc/issues/1661
    """
    builder = GlyphSetBuilder()
    # Set up font with two masters on the weight axis
    master0 = GSFontMaster()
    master0.id = "master-0"
    master0.axes = [100]  # light
    builder.font.masters.append(master0)
    master1 = GSFontMaster()
    master1.id = "master-1"
    master1.axes = [200]  # bold
    builder.font.masters.append(master1)

    # 'ae': base glyph with anchors at both masters
    builder.add_glyph(
        "ae",
        lambda glyph: (
            glyph.add_anchor("top", (300, 600))
            .add_anchor("bottom", (300, 0))
            .add_master_layer()
            .add_anchor("top", (380, 700))
            .add_anchor("bottom", (380, 0))
        ),
    )

    # 'acutecomb': mark glyph with anchors at both masters only (no brace layer)
    builder.add_glyph(
        "acutecomb",
        lambda glyph: (
            glyph.add_anchor("_top", (100, 500))
            .add_anchor("top", (120, 700))
            .add_master_layer()
            .add_anchor("_top", (140, 520))
            .add_anchor("top", (160, 740))
        ),
    )

    # 'aeacute': composite with 2 masters + a brace layer at weight=150
    builder.add_glyph(
        "aeacute",
        lambda glyph: (
            glyph.add_component("ae", (0, 0))
            .add_component("acutecomb", (200, 100))
            .add_master_layer()
            .add_component("ae", (0, 0))
            .add_component("acutecomb", (240, 120))
            # Add brace layer at weight=150 (midpoint)
            .add_brace_layer([150], associated_master_idx=0)
            .add_component("ae", (0, 0))
            .add_component("acutecomb", (220, 110))
        ),
    )

    propagate_all_anchors(builder.font)

    aeacute = builder.font.glyphs["aeacute"]

    # Master 0 (weight=100): acutecomb has explicit sources
    # top from acutecomb.top (120, 700) + offset (200, 100) = (320, 800)
    # bottom from ae: (300, 0)
    assert_anchors(
        aeacute.layers["master-0"].anchors,
        [("top", (320, 800)), ("bottom", (300, 0))],
    )

    # Master 1 (weight=200): acutecomb has explicit sources
    # top from acutecomb.top (160, 740) + offset (240, 120) = (400, 860)
    # bottom from ae: (380, 0)
    assert_anchors(
        aeacute.layers["master-1"].anchors,
        [("top", (400, 860)), ("bottom", (380, 0))],
    )

    # Brace layer (weight=150): both ae and acutecomb must be INTERPOLATED
    # Interpolated ae anchors at weight=150 (midpoint of 100..200):
    #   top: (300, 600) + 0.5 * ((380, 700) - (300, 600)) = (340, 650)
    #   bottom: (300, 0) + 0.5 * ((380, 0) - (300, 0)) = (340, 0)
    # Interpolated acutecomb anchors at weight=150:
    #   _top: (100, 500) + 0.5 * ((140, 520) - (100, 500)) = (120, 510)
    #   top:  (120, 700) + 0.5 * ((160, 740) - (120, 700)) = (140, 720)
    # Component offsets at brace layer: acutecomb offset = (220, 110)
    # Final top = interpolated acutecomb.top + offset
    #           = (140, 720) + (220, 110) = (360, 830)
    # Final bottom = interpolated ae.bottom = (340, 0)
    brace_layers = [l for l in aeacute.layers if l._is_brace_layer()]
    assert len(brace_layers) == 1
    assert_anchors(
        brace_layers[0].anchors,
        [("top", (360, 830)), ("bottom", (340, 0))],
    )


def test_brace_layer_matching():
    """When the component has a matching brace layer (same coordinates +
    associated master), use it directly without interpolation."""
    builder = GlyphSetBuilder()
    master0 = GSFontMaster()
    master0.id = "master-0"
    master0.axes = [100]
    builder.font.masters.append(master0)
    master1 = GSFontMaster()
    master1.id = "master-1"
    master1.axes = [200]
    builder.font.masters.append(master1)

    # 'base': has a brace layer at weight=150
    builder.add_glyph(
        "base",
        lambda glyph: (
            glyph.add_anchor("top", (100, 600))
            .add_master_layer()
            .add_anchor("top", (200, 700))
            .add_brace_layer([150], associated_master_idx=0)
            .add_anchor("top", (160, 660))  # explicit, not interpolated
        ),
    )

    # 'composite': also has a brace layer at weight=150
    builder.add_glyph(
        "composite",
        lambda glyph: (
            glyph.add_component("base", (0, 0))
            .add_master_layer()
            .add_component("base", (0, 0))
            .add_brace_layer([150], associated_master_idx=0)
            .add_component("base", (10, 0))
        ),
    )

    propagate_all_anchors(builder.font)

    composite = builder.font.glyphs["composite"]
    brace_layers = [l for l in composite.layers if l._is_brace_layer()]
    assert len(brace_layers) == 1
    # Should use the component's explicit brace layer anchor (160, 660) + offset (10, 0)
    assert_anchors(brace_layers[0].anchors, [("top", (170, 660))])


def test_brace_layer_interpolation_with_existing_brace_sources():
    """Component has brace layers at different coordinates; interpolation uses
    both masters and existing brace layers as sources."""
    builder = GlyphSetBuilder()
    master0 = GSFontMaster()
    master0.id = "master-0"
    master0.axes = [100]
    builder.font.masters.append(master0)
    master1 = GSFontMaster()
    master1.id = "master-1"
    master1.axes = [300]
    builder.font.masters.append(master1)

    # 'base': 2 masters + a brace layer at weight=200
    builder.add_glyph(
        "base",
        lambda glyph: (
            glyph.add_anchor("top", (100, 600))
            .add_master_layer()
            .add_anchor("top", (300, 800))
            .add_brace_layer([200], associated_master_idx=0)
            .add_anchor("top", (220, 720))  # off the linear path
        ),
    )

    # 'composite': brace layer at weight=150 (not matched by component)
    builder.add_glyph(
        "composite",
        lambda glyph: (
            glyph.add_component("base", (0, 0))
            .add_master_layer()
            .add_component("base", (0, 0))
            .add_brace_layer([150], associated_master_idx=0)
            .add_component("base", (0, 0))
        ),
    )

    propagate_all_anchors(builder.font)

    composite = builder.font.glyphs["composite"]
    brace_layers = [l for l in composite.layers if l._is_brace_layer()]
    assert len(brace_layers) == 1

    # The component has 3 sources: weight=100, 200, 300.
    # Interpolating at weight=150 with a 3-source model should use the brace
    # layer at 200 as well, giving a different result than simple 2-master lerp.
    # With 2-master linear:
    #   top = (100, 600) + 0.25 * ((300, 800) - (100, 600)) = (150, 650)
    # With 3-source model the result differs because (220, 720) at weight=200 is
    # off the linear path between (100, 600) and (300, 800).
    anchors = brace_layers[0].anchors
    assert len(anchors) == 1
    assert anchors[0].name == "top"
    # Verify it's NOT the 2-master linear result
    assert (round(anchors[0].position.x), round(anchors[0].position.y)) != (150, 650)


def test_interpolate_brace_uses_default_source_anchor_names():
    """Canonical anchor names should come from the default source location.

    If an anchor is present in the default source but missing from another source,
    it must still be interpolated (sparse interpolation), not dropped.
    """
    builder = GlyphSetBuilder()
    master0 = GSFontMaster()
    master0.id = "master-0"
    master0.name = "Bold"
    master0.axes = [200]
    builder.font.masters.append(master0)
    master1 = GSFontMaster()
    master1.id = "master-1"
    master1.name = "Regular"
    master1.axes = [100]
    builder.font.masters.append(master1)
    # Make default source != first source layer
    builder.font.customParameters["Variable Font Origin"] = "master-1"

    # Component:
    # - non-default source (master-0): top only
    # - default source (master-1): top + bottom
    builder.add_glyph(
        "base",
        lambda glyph: (
            glyph.add_anchor("top", (200, 700))
            .add_master_layer()
            .add_anchor("top", (100, 600))
            .add_anchor("bottom", (100, 0))
        ),
    )

    # Composite has a brace layer at weight=150; component has no matching brace layer
    builder.add_glyph(
        "comp",
        lambda glyph: (
            glyph.add_component("base", (0, 0))
            .add_master_layer()
            .add_component("base", (0, 0))
            .add_brace_layer([150], associated_master_idx=0)
            .add_component("base", (0, 0))
        ),
    )

    propagate_all_anchors(builder.font)

    comp = builder.font.glyphs["comp"]
    brace_layers = [l for l in comp.layers if l._is_brace_layer()]
    assert len(brace_layers) == 1
    # `bottom` must not be dropped just because it's missing in non-default source.
    assert_anchors(
        brace_layers[0].anchors,
        [("top", (150, 650)), ("bottom", (100, 0))],
    )
